from . import (xmlrpc,   # noqa
               jsonrpc,  # noqa
               local)    # noqa
from .connection import (ConnectorBase,        # noqa
                         get_connector,        # noqa
                         get_connector_names)  # noqa



import six
from extend_me import ExtensibleByHashType

__all__ = ('get_connector', 'get_connector_names', 'ConnectorBase')

ConnectorType = ExtensibleByHashType._('Connector', hashattr='name')


def get_connector(name):
    """ Return connector specified by it's name
    """
    return ConnectorType.get_class(name)


def get_connector_names():
    """ Returns list of connector names registered in system
    """
    return ConnectorType.get_registered_names()


class ConnectorBase(six.with_metaclass(ConnectorType)):
    """ Base class for all connectors

        :param str host: hostname to connect to
        :param int port: port to connect to
        :param dict extra_args: extra arguments for specific connector.
    """

    def __init__(self, host, port, extra_args=None):
        self._host = host
        self._port = port
        self._extra_args = {} if extra_args is None else extra_args

        self.__services = {}

    @property
    def host(self):
        """ Connector host
        """
        return self._host

    @property
    def port(self):
        """ Connector port
        """
        return self._port

    @property
    def extra_args(self):
        """ Connector extra arguments
        """
        return self._extra_args

    def update_extra_args(self, **kwargs):
        """ Update extra args and clean service cache
        """
        self.extra_args.update(kwargs)
        self.__services = {}

    def _get_service(self, name):  # pragma: no cover
        raise NotImplementedError

    def get_service(self, name):
        """ Returns service for specified *name*

            :param name: name of service
            :return: specified service instance
        """
        service = self.__services.get(name, None)
        if service is None:
            service = self._get_service(name)
            self.__services[name] = service

        return service



import six
from extend_me import ExtensibleByHashType
from pkg_resources import parse_version

from ..utils import (AttrDict,
                     DirMixIn,
                     preprocess_args,
                     stdcall)


__all__ = ('Object', 'get_object')


ObjectType = ExtensibleByHashType._('Object', hashattr='name')


def get_object(client, name):
    """ Create new Object instance.

        :param client: Client instance to bind this object to
        :type client: Client
        :param name: name of object. Ex. 'sale.order'
        :type name: str
        :return: Created Object instance
        :rtype: Object
    """
    cls = ObjectType.get_class(name, default=True)
    return cls(client, name)


# TODO: implement clean caches new columns may be defined, when new addon was
# installed
@six.python_2_unicode_compatible
class Object(six.with_metaclass(ObjectType, DirMixIn)):
    """ Base class for all Objects

        Provides simple interface to remote osv.osv objects::

            erp = Client(...)
            sale_obj = Object(erp, 'sale.order')
            sale_obj.search([('state','not in',['done','cancel'])])

        To create new instance - use *get_object* function, it implements
        all extensions magic, whic is highly used in this project

        It is posible to create extension only to specific object.
        Example could be found in ``plugins/module_utils.py`` file.

    """

    __slots__ = ('_service', '_obj_name', '_columns_info')

    def __init__(self, service, object_name):
        self._service = service
        self._obj_name = object_name

        self._columns_info = None

    @property
    def name(self):
        """ Name of the object

            :rtype: str
        """
        return self._obj_name

    @property
    def service(self):
        """ Object service instance
        """
        return self._service

    @property
    def client(self):
        """ Client instance, this object is relatedto

            :rtype: odoo_rpc_client.client.Client
        """
        return self.service.client

    def __getattr__(self, name):
        def method_wrapper(object_name, method_name):
            """ Wraper around Odoo objects's methods.

                for internal use.
                It is used in Object class.
            """
            @stdcall
            def wrapper(*args, **kwargs):
                return self.service.execute(object_name,
                                            method_name,
                                            *args,
                                            **kwargs)
            name = str('%s:%s' % (object_name, method_name))
            wrapper.__name__ = name
            return wrapper

        # Private methods are not available to be called via RPC
        if name.startswith('_'):
            raise AttributeError("Private methods are not exposed to RPC. "
                                 "(attr: %s)" % name)

        setattr(self, name,  method_wrapper(self.name, name))
        return getattr(self, name)

    def __str__(self):
        return u"Object ('%s')" % self.name

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        assert isinstance(other, Object), \
            "Comparable only with instances of Object class"
        return self.name == other.name and self.client == other.client

    def _get_columns_info(self):
        """ Calculates columns info
        """
        return AttrDict(self.fields_get())

    @property
    def columns_info(self):
        """ Reads information about fields available on model
        """
        if self._columns_info is None:
            self._columns_info = self._get_columns_info()

        return self._columns_info

    def resolve_field_path(self, field):
        """ Resolves dot-separated field path
            to list of tuples (model, field_name, related_model)

            :param str field: dot-separated field path to resolve

            For example::

                sale_obj = client['sale.order']
                sale_obj.resolve_field_path('partner_id.country_id.name')

            will be resoved to::

                [('sale.order', 'partner_id', 'res.partner'),
                 ('res.partner', 'country_id', 'res.country'),
                 ('res.country', 'name', False)]

        """
        field_path = field.split('.')
        res = []

        model = self.name
        cinfo = self.client[model].columns_info
        f = field_path.pop(0)
        res.append((model, f, cinfo[f].get('relation', False)))

        while field_path:
            model = cinfo[f]['relation']
            cinfo = self.client[model].columns_info
            f = field_path.pop(0)
            res.append((model, f, cinfo[f].get('relation', False)))
        return res

    @property
    def stdcall_methods(self):
        """ Property that returns all methods of this object,
            that supports standard call

            :return: list with names of *stdcall* methods
            :rtype: list(str)
        """
        return [n for n in dir(self)
                if not n.startswith('_') and
                n != 'stdcall_methods' and
                getattr(getattr(self, n, None), '__x_stdcall__', False)]

    @stdcall
    def read(self, ids, fields=None, context=None):
        """ Read *fields* for records with id in *ids*

            Also look at `Odoo documentation
            <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.read>`__
            for this method

            :param int|list ids: ID or list of IDs of records to read data for
            :param list fields: list of field names to read.
                                if not passed all fields will be read.
            :param dict context: dictionary with extra context
            :return: list of dictionaries with data had been read
            :rtype: list
        """
        args, kwargs = preprocess_args(ids, fields, context=context)
        res = self.service.execute(self.name, 'read', *args, **kwargs)

        # Odoo 10 compatability fix.
        # In Odoo 10, ``model.read(<id>)`` will return list, NOT dict as in
        # previous Odoo versions. So fix it to make api compatible with older
        # versions
        if res and isinstance(ids, int) and isinstance(res, list):
            res = res[0]
        return res

    @stdcall
    def write(self, ids, vals, context=None):
        """ Write data in *vals* dictionary to records with ID in *ids*

            For more info, look at `odoo documentation <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.write>`__
            for this method

            :param int|list ids: ID or list of IDs of records to write data for
            :param dict vals: dictinary with values to be written to database
                              for records specified by ids
            :param dict context: context dictionary
        """  # noqa
        args, kwargs = preprocess_args(ids, vals, context=context)
        return self.service.execute(self.name, 'write', *args, **kwargs)

    def create(self, vals, context=None):
        """ Create new record with *vals*

            Also look at `Odoo documentation <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.create>`__
            for this method

            :param dict vals: dictionary with values to be written
                              to newly created record
            :param dict context: context dictionary
            :return: ID of newly created record
            :rtype: int
        """  # noqa
        args, kwargs = preprocess_args(vals, context=context)
        return self.service.execute(self.name, 'create', *args, **kwargs)

    @stdcall
    def unlink(self, ids, context=None):
        """ Unlink records specified by *ids*

            Also look at `Odoo documentation <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.unlink>`__
            for this method

            :param list ids: list of IDs of records to be deleted
        """  # noqa
        args, kwargs = preprocess_args(ids, context=context)
        return self.service.execute(self.name, 'unlink', *args, **kwargs)

    def search(self, *args, **kwargs):
        """search(args[, offset=0][, limit=None][, order=None][, count=False][, context=None])

            Search records by criteria.

            Also look at `Odoo documentation <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.search>`__
            for this method
        """  # noqa
        _, kwargs = preprocess_args(**kwargs)  # preprocess kwargs
        return self.service.execute(self.name, 'search', *args, **kwargs)

    def search_read(self, domain=None, fields=None, offset=0, limit=None,
                    order=None, context=None):
        """ Search and read records specified by domain

            Note that this method reads data in correct order

            Also look at `Odoo documentation <https://www.odoo.com/documentation/9.0/reference/orm.html#openerp.models.Model.search_read>`__

            :return: list of dictionaries with data had been read
            :rtype: list
        """  # noqa
        if self.client.server_version >= parse_version('8.0'):
            args, kwargs = preprocess_args(domain=domain,
                                           fields=fields,
                                           offset=offset,
                                           limit=limit,
                                           order=order,
                                           context=context)
            return self.service.execute(self.name, 'search_read', **kwargs)
        else:
            ids = self.search(domain,
                              offset=offset,
                              limit=limit,
                              order=order,
                              context=context)
            read = self.read(ids, fields=fields, context=context)

            # reorder read
            index = dict((r['id'], r) for r in read)
            return [index[x] for x in ids if x in index]



import six
import re
from pkg_resources import parse_version

from ..service.service import ServiceBase

__all__ = ('DBService',)


def to_dbname(db):
    """ Converts db to string, that represents database name

        :param db: Client instance or string with name of database
        :type db: Client|str
        :return: string with database name
        :rtype: str
        :raises ValueError: value of db is not parsable
    """
    from odoo_rpc_client.client import Client
    if isinstance(db, six.string_types):
        return db
    elif isinstance(db, Client) and db.dbname is not None:
        return db.dbname
    else:
        raise ValueError("Wrong value for db!")


class DBService(ServiceBase):
    """ Service class to simplify interaction with 'db' service
    """
    class Meta:
        name = 'db'

    def list_db(self):
        """ Display list of databses of thist connection
        """
        return self._service.list()

    def db_exist(self, db):
        """ Check if database exists

            :param str|Client db: name of database or *Client* instance
                                  with *client.dbname is not None*
            :return: True if database exists else False
            :rtype: bool
        """
        return self._service.db_exist(to_dbname(db))

    def create_db(self, password, dbname, demo=False, lang='en_US',
                  admin_password='admin'):
        """ Create new database on server, named *dbname*

            :param str password: super admin password
            :param str dbname: name of database to create
            :param bool demo: load demo data or not. Default: False
            :param str lang: language to be used for database. Default: 'en_US'
            :param str admin_password: password to be used for 'Administrator'
                                       database user.
                                       Default: 'admin'
            :return: Client instance logged to created database as admin user.
            :rtype: instance of *odoo_rpc_client.client.Client*
        """
        from odoo_rpc_client.client import Client

        self.create_database(password, dbname, demo, lang, admin_password)

        client = Client(self.client.host, port=self.client.port,
                        protocol=self.client.protocol, dbname=dbname,
                        user='admin', pwd=admin_password)
        return client

    def drop_db(self, password, db):
        """ Drop specified database

            :param str password: super admin password
            :param str|Client db: name of database or *Client* instance
                                  with *client.dbname is not None*
            :raise: `ValueError` (unsupported value of *db* argument)
        """
        return self.drop(password, to_dbname(db))

    def dump_db(self, password, db, **kwargs):
        """ Dump database

            Note, that from defined arguments, may be passed other arguments
            (for example odoo version 9.0 requires format arg to be passed)

            :param str password: super admin password
            :param str|Client db: name of database or *Client* instance
                                  with *client.dbname is not None*
            :param str format: (only odoo 9.0) (default: zip)
            :raise: `ValueError` (unsupported value of *db* argument)
            :return: byte-string with base64 encoded data
            :rtype: bytes
        """
        # format argument available only for odoo version 9.0
        if self.server_base_version() >= parse_version('9.0'):
            args = [kwargs.get('format', 'zip')]
        else:
            args = []

        dump_data = self.dump(password, to_dbname(db), *args)

        # Ensure returned data is not unicode, but bytes
        if isinstance(dump_data, six.text_type):
            dump_data = dump_data.encode()

        return dump_data

    def restore_db(self, password, dbname, data, **kwargs):
        """ Restore database

            :param str password: super admin password
            :param str dbname: name of database
            :param bytes data: restore data (base64 encoded string)
            :param bool copy: (only odoo 8.0+) if set to True,
                              then new db-uid will be generated.
                              (default: False)
            :return: True
            :rtype: bool
        """
        assert isinstance(data, bytes), \
            "data must be instance of bytes. got: %s" % type(data)
        if self.server_base_version() >= parse_version('8.0') and \
                'copy' in kwargs:
            args = [kwargs['copy']]
        else:
            args = []

        # data passed to restore should be Unicode, otherwise it dumps it as
        # binary data on python3
        if six.PY3 and isinstance(data, six.binary_type):
            data = data.decode()

        return self.restore(password, dbname, data, *args)

    def server_version(self):
        """ Returns server version.

            (Already parsed with pkg_resources.parse_version)
        """
        return parse_version(self.server_version_str())

    def server_base_version(self):
        """ Returns server base version ('9.0', '8.0', etc)
            parsed via pkg_resources.parse_version.
            No info about comunity / enterprise here
        """
        base_version = self.server_version().base_version

        # Remove 'rc.*' suffix
        base_version = re.sub(r'rc.+', '', base_version)
        return parse_version(base_version)

    def server_version_str(self):
        """ Return server version (not wrapped by pkg.parse_version)
        """
        return self._service.server_version()

    # make able to check if there some databases on server exists
    def __contains__(self, name):
        return name in self.list_db()



# -*- coding: utf8 -*-
"""
This module provides some classes to simplify access to Odoo server via xmlrpc.

Example ussage of this module

.. code:: python

    >>> cl = Client('server.com', 'dbname', 'some_user', 'mypassword')
    >>> sale_obj = cl['sale_order']
    >>> sale_ids = sale_obj.search([('state','not in',['done','cancel'])])
    >>> sale_data = sale_obj.read(sale_ids, ['name'])
    >>> for order in sale_data:
    ...     print("%5s :    %s" % (order['id'],order['name']))
    >>> product_tmpl_obj = cl['product.template']
    >>> product_obj = cl['product.product']
    >>> tmpl_ids = product_tmpl_obj.search([('name','ilike','template_name')])
    >>> print(product_obj.search([('product_tmpl_id','in',tmpl_ids)]))

    >>> db = Client('erp.host.com', 'dbname='db0', user='your_user')
    >>> so = db['sale.order']
    >>> order_ids = so.search([('state','=','done')])
    >>> order = so.read(order_ids[0])

Also You can call any method (beside private
ones starting with underscore(_)) of any model.
For example following code allows to check
availability of stock moves:

.. code:: python

    >>> db = session.connect()
    >>> move_obj = db['stock.move']
    >>> move_ids = [1234] # IDs of stock moves to be checked
    >>> move_obj.check_assign(move_ids)

Ability to use Record class as analog to browse_record:

.. code:: python

    >>> move_obj = db['stock.move']
    >>> move = move_obj.browse(1234)
    >>> move.state
    ... 'confirmed'
    >>> move.check_assign()
    >>> move.refresh()
    >>> move.state
    ... 'assigned'
    >>> move.picking_id
    ... R('stock.picking', 12)['OUT-12']
    >>> move.picking_id.id
    ... 12
    >>> move.picking_id.name
    ... 'OUT-12'
    >>> move.picking_id_.state
    ... 'assigned'
"""

import six
import re
from extend_me import Extensible
from pkg_resources import parse_version

# project imports
from .connection import get_connector
from .exceptions import LoginException
from .service import ServiceManager
from .plugin import PluginManager

# Enable ORM features
from . import orm  # noqa


__all__ = ('Client',)

RE_CLIENT_URL = re.compile(
    r"(?:(?P<protocol>[\w\-]+)\:\/\/)?(?:(?P<user>[\w\-]+)?"
    r"(?:\:(?P<pwd>[\w\-\.\,]+))?\@)?"
    r"(?P<host>[\w\-\.]+)(?:\:(?P<port>\d{2,5}))?\/"
    r"(?P<dbname>[\w\-.]+)?$")


@six.python_2_unicode_compatible
class Client(Extensible):
    """
       A simple class to connect to Odoo instance via RPC (XML-RPC, JSON-RPC)
       Should be initialized with following arguments:

       :param str host: server host name to connect to
       :param str dbname: name of database to connect to
       :param str user: username to login as
       :param str pwd: password to log-in with
       :param int port: port number of server
       :param str protocol: protocol used to connect.
                            To get list of available protcols call:
                            ``odoo_rpc_client.connection.get_connector_names()``

       any other keyword arguments will be directly passed to connector

       Example::

           >>> db = Client('host', 'dbname', 'user', pwd = 'Password')
           >>> cl = Client('host')
           >>> db2 = cl.login('dbname', 'user', 'password')

       Allows access to Odoo objects / models via dictionary syntax::

           >>> db['sale.order']
               Object ('sale.order')
    """

    def __init__(self, host, dbname=None, user=None, pwd=None, port=8069,
                 protocol='xml-rpc', **extra_args):
        self._dbname = dbname
        self._username = user
        self._pwd = pwd

        self._connection = get_connector(protocol)(host, port, extra_args)
        self._services = ServiceManager(self)
        self._plugins = PluginManager(self)

        self._uid = None
        self._user = None
        self._user_context = None
        self._database_version_full = None

    @property
    def dbname(self):
        """ Name of database to connect to

            :rtype: str
        """
        return self._dbname

    @property
    def username(self):
        """ User login used to access DB

            :rtype: str
        """
        return self._username

    @property
    def host(self):
        """ Server host

            :rtype: str
        """
        return self._connection.host

    @property
    def port(self):
        """ Server port
        """
        return self._connection.port

    @property
    def protocol(self):
        """ Server protocol

            :rtype: str
        """
        return self._connection.Meta.name

    @property
    def services(self):
        """ ServiceManager instance, which contains list
            of all available services for current connection.

            :rtype: odoo_rpc_client.service.service.ServiceManager

            Usage examples::

                db.services.report   # report service
                db.services.object   # object service (model related actions)
                db.services.common   # used for login
                                     # (db.services.common.login(dbname,
                                     #                           username,
                                     #                           password)
                db.services.db       # database management service

        """
        return self._services

    @property
    def plugins(self):
        """ Plugins associated with this Client instance

            :rtype: odoo_rpc_client.plugin.PluginManager

            Usage examples::

                db.plugins.module_utils    # access module_utils plugin
                db.plugins['module_utils]  # access module_utils plugin

        """
        return self._plugins

    @property
    def connection(self):
        """ Connection to server.

            :rtype: odoo_rpc_client.connection.connection.ConnectorBase
        """
        return self._connection

    @property
    def uid(self):
        """ Returns ID of current user. if one is None,
            connects to database and returns it

            :rtype: int
        """
        if self._uid is None:
            self._uid = self.connect()
        return self._uid

    @property
    def user(self):
        """ Currenct logged in user instance

            :rtype: odoo_rpc_client.orm.record.Record
        """
        if self._user is None:
            self._user = self.get_obj('res.users').read_records(self.uid)
        return self._user

    @property
    def user_context(self):
        """ Get current user context

            :rtype: dict
        """
        if self._user_context is None:
            self._user_context = self.get_obj('res.users').context_get()
        return self._user_context

    @property
    def server_version(self):
        """ Server base version  ('8.0', '9.0', etc)

            (Already parsed with ``pkg_resources.parse_version``)
        """
        return self.services.db.server_base_version()

    @property
    def database_version_full(self):
        """ Full database base version ('9.0.1.3', etc)

            (Already parsed with ``pkg_resources.parse_version``)
        """
        if self._database_version_full is None:
            base_module = self.get_obj('ir.module.module').search_records(
                [('name', '=', 'base')])[0]
            self._database_version_full = parse_version(
                base_module.installed_version)
        return self._database_version_full

    @property
    def database_version(self):
        """ Base database version ('8.0', '9.0', etc)

            (Already parsed with ``pkg_resources.parse_version``)
        """
        return parse_version(
            '.'.join(
                self.database_version_full.base_version.split('.', 2)[:2]))

    @property
    def registered_objects(self):
        """ List of registered in Odoo database objects

            :rtype: list
        """
        return self.services['object'].get_registered_objects()

    def login(self, dbname, user, password):
        """ Login to database

            Return new Client instance.
            (Just an aliase on ``connect`` method)

            :param str dbname: name of database to connect to
            :param str user: username to login as
            :param str password: password to log-in with
            :return: new Client instance, with specifed credentials
            :rtype: odoo_rpc_client.client.Client
        """
        return self.connect(dbname=dbname, user=user, pwd=password)

    def connect(self, **kwargs):
        """ Connects to the server

            if any keyword arguments will be passed, new Proxy instnace
            will be created using folowing algorithm: get init args from
            self instance and update them with passed keyword arguments,
            and call Proxy class constructor passing result as arguments.

            **Note**, that if You pass any keyword arguments,
            You also should pass 'pwd' keyword argument with user password

            :return: Id of user logged in or new Client
                     instance (if kwargs passed)
            :rtype: int|Client
            :raises LoginException: if wrong login or password
        """
        if kwargs:
            init_kwargs = self.get_init_args()
            init_kwargs.update(kwargs)

            return Client(**init_kwargs)

        # Get the uid
        if not self._pwd or not self.username or not self.dbname:
            raise LoginException("User login and password and dbname required "
                                 "for this operation")

        uid = self.services['common'].login(self.dbname,
                                            self.username,
                                            self._pwd)

        if not uid:
            raise LoginException("Bad login or password")

        return uid

    def reconnect(self):
        """ Recreates connection to the server and clears caches

            :return: ID of user logged in
            :rtype: int
            :raises ClientException: if wrong login or password
        """
        self.services.clean_cache()
        self._uid = None
        self._uid = self.connect()
        return self._uid

    def execute(self, obj, method, *args, **kwargs):
        """Call method *method* on object *obj* passing all next
           positional and keyword (if available on server)
           arguments to remote method

           Note that passing keyword argments not available on
           OpenERp/Odoo server 6.0 and older

           :param obj: object name to call method for
           :type obj: string
           :param method: name of method to call
           :type method: string
           :return: result of RPC method call
        """
        return self.services['object'].execute(obj, method, *args, **kwargs)

    def execute_wkf(self, object_name, signal, object_id):
        """ Triggers workflow event on specified object

            :param object_name: send workflow signal for
            :type object_name: string
            :param signal: name of signal to send
            :type signal: string
            :param object_id: ID of document (record) to send signal to
            :type obejct_id: int
        """
        result_wkf = self.services['object'].execute_wkf(object_name,
                                                         signal,
                                                         object_id)
        return result_wkf

    def get_obj(self, object_name):
        """ Returns wraper around Odoo object 'object_name'
            which is instance of orm.object.Object class

            :param object_name: name of an object to get wraper for
            :return: instance of Object which wraps choosen object
            :rtype: odoo_rpc_client.orm.object.Object
        """
        return self.services['object'].get_obj(object_name)

    def ref(self, xmlid):
        """ Return record for specified xmlid

            :param str xmlid: string representing xmlid to get record for.
                              xmlid must be *fully qualified*
                              (with module name)
            :return: Record for that xmlid or False
            :rtype: odoo_rpc_client.orm.record.Record
        """
        try:
            module, name = xmlid.split('.')
        except ValueError:
            raise ValueError(
                "Fully qualified xmlid required! (Ex. 'module_name.xmlid'")
        res = self['ir.model.data'].search_records(
            [('module', '=', module), ('name', '=', name)],
            limit=1)
        if res:
            res = res[0]
            return self[res.model].read_records(res.res_id)

        return False

    def __getitem__(self, name):
        """ Returns instance of Object with name 'name'
        """
        res = None
        try:
            res = self.get_obj(name)
        except ValueError:
            raise KeyError('Wrong object/model name: %s' % name)

        return res

    def get_init_args(self):
        """ Returns dictionary with init arguments which can be safely passed
            to class constructor

            :rtype: dict
        """
        return dict(user=self.username,
                    host=self.host,
                    port=self.port,
                    dbname=self.dbname,
                    protocol=self.protocol,
                    **self.connection.extra_args)

    @classmethod
    def to_url(cls, inst, **kwargs):
        """ Converts instance to url

            :param inst: instance to convert to init args
            :type inst: Client|dict
            :return: generated URL
            :rtype: str
        """
        url_tmpl = "%(protocol)s://%(user)s@%(host)s:%(port)s/%(dbname)s"
        if isinstance(inst, Client):
            return url_tmpl % inst.get_init_args()
        elif isinstance(inst, dict):
            return url_tmpl % inst
        elif inst is None and kwargs:
            return url_tmpl % kwargs
        else:
            raise ValueError("inst must be Client instance or dict")

    @classmethod
    def from_url(cls, url):
        """ Create Client instance from URL

            :param str url: url of Client
            :return: Client instance
            :rtype: Client
        """
        m = RE_CLIENT_URL.match(url)
        if m:
            data = dict(m.groupdict())
            data['protocol'] = data.get('protocol', None) or 'xml-rpc'
            data['port'] = int(data.get('port', None) or '80')
            return Client(**data)
        raise ValueError("Cannot parse url")

    # TODO: think to reimplement as property
    def get_url(self):
        """ Returns dabase URL

            At this moment mostly used internaly in session
        """
        return self.to_url(self)

    def clean_caches(self):
        """ Clean client related caches
        """
        self.services.clean_service_caches()
        self.plugins.refresh()
        self._user_context = None
        self._user = None
        self._database_version_full = None

    def __str__(self):
        return u"Client: %s" % self.get_url()

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, Client):
            return self.get_url() == other.get_url()
        else:
            return False



from . import BaseTestCase
from ..client import Client
from ..exceptions import (LoginException,
                          ConnectorError)
from ..connection import get_connector_names


class Test_00_Connection(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             protocol=self.env.protocol,
                             port=self.env.port)

    def test_00_connection_simple(self):
        client = self.client
        self.assertIsNone(client.username)
        self.assertIsNone(client.dbname)

        # access to uid with out credentials provided, should cause error
        with self.assertRaises(LoginException):
            client.uid

        # access to user with out credentials provided, should cause error
        with self.assertRaises(LoginException):
            client.user

    def test_01_create_db(self):
        client = self.client
        if self.env.dbname in client.services.db.list_db():
            self.assertIn(self.env.dbname, client.services.db)

            if self.recreate_db:
                client.services.db.drop_db(self.env.super_password,
                                           self.env.dbname)
            else:
                return self.skipTest("Database already created")

        self.assertNotIn(self.env.dbname, client.services.db)

        cl = client.services.db.create_db(self.env.super_password,
                                          self.env.dbname,
                                          demo=True,
                                          admin_password=self.env.password)

        self.assertIn(self.env.dbname, client.services.db)
        # cl is object of same class as client, but with credential filled
        self.assertIsInstance(cl, Client)

        # test that uid and user properties are accessible
        self.assertIsNotNone(cl.uid)
        self.assertIsNotNone(cl.user)

    def test_02_connect(self):
        client = self.client
        cl = client.connect(dbname=self.env.dbname,
                            user=self.env.user,
                            pwd=self.env.password)

        # cl is object of same class as client, but with credential filled
        self.assertIsInstance(cl, Client)

        # test that uid and user properties are accessible
        self.assertIsNotNone(cl.uid)
        self.assertIsNotNone(cl.user)

    def test_03_login(self):
        client = self.client
        cl = client.login(self.env.dbname, self.env.user, self.env.password)

        # cl is object of same class as client, but with credential filled
        self.assertIsInstance(cl, Client)

        # test that uid and user properties are accessible
        self.assertIsNotNone(cl.uid)
        self.assertIsNotNone(cl.user)

    def test_04_login_wrong(self):
        client = self.client

        with self.assertRaises(LoginException):
            wrong_password = self.env.password + "-wrong"
            cl = client.login(self.env.dbname, self.env.user, wrong_password)
            # Note, that login return's new instance of client,
            # and we need to acces uid property,
            # to make it try to login via RPC.
            cl.uid

    def test_05_reconnect(self):
        client = self.client
        cl = client.login(self.env.dbname, self.env.user, self.env.password)
        old_uid = cl.uid

        new_uid = cl.reconnect()
        self.assertEqual(old_uid, new_uid)
        self.assertEqual(old_uid, cl.uid)

    def test_06_get_connector_names(self):
        self.assertItemsEqual(
            get_connector_names(),
            ['json-rpc', 'json-rpcs', 'xml-rpc', 'xml-rpcs', 'local'])

    def test_10_call_unexistint_method(self):
        cl = self.client.login(self.env.dbname,
                               self.env.user,
                               self.env.password)
        with self.assertRaises(ConnectorError):
            cl.execute('res.partner', 'some_unexistent_method__42')

    def test_15_call_unexistent_service_method(self):
        cl = self.client.login(self.env.dbname,
                               self.env.user,
                               self.env.password)
        with self.assertRaises(ConnectorError):
            cl.services['unexistent_service_42'].call_unexistent_method_78()



# -*- coding: utf-8 -*-

import six

from . import BaseTestCase
from .. import utils


class Test_100_Utils_ustr(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()

        # Note, if cp1251 will be first, than tests will fail
        # because following code will not throw errors:
        #    u'Юнікод'.encode('utf8').decode('cp1251')
        self.ustr2 = utils.UConverter(['utf8', 'cp1251', 'ascii'])

        class UObj(object):
            if six.PY3:
                def __str__(self):
                    return u"Юнікод"
            elif six.PY2:
                def __unicode__(self):
                    return u"Юнікод"

        class BObj(object):
            if six.PY3:
                # TODO: Is this correct for python3
                # for python 3 repr will be called
                def __str__(self):
                    return u"Юнікод"
            elif six.PY2:
                def __str__(self):
                    return u"Юнікод".encode('utf-8')

        self.uobj = UObj()
        self.bobj = BObj()

    def test_ustr_unicode(self):
        ustr = utils.ustr

        self.assertIsInstance(ustr(u"Юнікод"), six.text_type)
        self.assertIs(ustr(u"Юнікод"), u"Юнікод")

    def test_ustr_bytes(self):
        ustr = utils.ustr

        self.assertIsInstance(u"Юнікод".encode("utf-8"), six.binary_type)
        self.assertIsInstance(ustr(u"Юнікод".encode("utf-8")), six.text_type)
        self.assertEqual(ustr(u"Юнікод".encode("utf-8")), u"Юнікод")

    def test_ustr_bytes_cp1251(self):
        ustr = utils.ustr

        # cp1251 encoding is not recognized by ustr (only utf8 and ascii)
        self.assertIsInstance(u"Юнікод".encode("cp1251"), six.binary_type)
        with self.assertRaises(UnicodeError):
            self.assertIsInstance(ustr(u"Юнікод".encode("cp1251")),
                                  six.text_type)
            self.assertEqual(ustr(u"Юнікод".encode("cp1251")), u"Юнікод")

    def test_ustr_uobj(self):
        ustr = utils.ustr

        self.assertIsInstance(ustr(self.uobj), six.text_type)
        self.assertEqual(ustr(self.uobj), u"Юнікод")

    def test_ustr_bobj(self):
        ustr = utils.ustr

        self.assertIsInstance(ustr(self.bobj), six.text_type)
        self.assertEqual(ustr(self.bobj), u"Юнікод")

    def test_ustr2_unicode(self):
        ustr = self.ustr2

        self.assertIsInstance(ustr(u"Юнікод"), six.text_type)
        self.assertIs(ustr(u"Юнікод"), u"Юнікод")

    def test_ustr2_bytes(self):
        ustr = self.ustr2

        self.assertIsInstance(u"Юнікод".encode("utf-8"), six.binary_type)
        self.assertIsInstance(ustr(u"Юнікод".encode("utf-8")), six.text_type)
        self.assertEqual(ustr(u"Юнікод".encode("utf-8")), u"Юнікод")

    def test_ustr2_bytes_cp1251(self):
        ustr = self.ustr2

        self.assertIsInstance(u"Юнікод".encode("cp1251"), six.binary_type)
        self.assertIsInstance(ustr(u"Юнікод".encode("cp1251")), six.text_type)
        self.assertEqual(ustr(u"Юнікод".encode("cp1251")), u"Юнікод")

    def test_ustr2_uobj(self):
        ustr = self.ustr2

        self.assertIsInstance(ustr(self.uobj), six.text_type)
        self.assertEqual(ustr(self.uobj), u"Юнікод")

    def test_ustr2_bobj(self):
        ustr = self.ustr2

        self.assertIsInstance(ustr(self.bobj), six.text_type)
        self.assertEqual(ustr(self.bobj), u"Юнікод")


class Test_102_Utils_AttrDict(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.attr_dict = utils.AttrDict(a=42, b=78)

    def test_isinstance(self):
        self.assertIsInstance(self.attr_dict, dict)
        self.assertIsInstance(self.attr_dict, utils.AttrDict)

    def test_getattr(self):
        self.assertIs(self.attr_dict.a, 42)
        self.assertIs(self.attr_dict.b, 78)

        with self.assertRaises(AttributeError):
            self.attr_dict.c

    def test_getitem(self):
        self.assertIs(self.attr_dict['a'], 42)
        self.assertIs(self.attr_dict['b'], 78)

        with self.assertRaises(KeyError):
            self.attr_dict['c']

    def test_dir(self):
        self.assertIn('a', dir(self.attr_dict))
        self.assertIn('b', dir(self.attr_dict))
        self.assertNotIn('c', dir(self.attr_dict))



from .test_connection import *      # noqa
from .test_client import *          # noqa
from .test_orm import *             # noqa
from .test_plugins import *         # noqa
from .test_utils import *           # noqa
from .test_service_report import *  # noqa
from .test_db import *              # noqa



from . import BaseTestCase
from .. import Client
from ..orm import (Record,
                   RecordList)

import unittest


class Test_25_Plugin_ModuleUtils(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)

    def test_10_init_module_utils(self):
        self.assertNotIn('module_utils', self.client.plugins)
        self.assertEqual(len(self.client.plugins), 1)
        self.assertIn('Test', self.client.plugins)

        import odoo_rpc_client.plugins.module_utils  # noqa

        self.assertIn('module_utils', self.client.plugins)
        self.assertEqual(len(self.client.plugins), 2)

    def test_15_modules(self):
        self.assertIsInstance(self.client.plugins.module_utils.modules, dict)
        self.assertIn('sale', self.client.plugins.module_utils.modules)
        self.assertIsInstance(self.client.plugins.module_utils.modules['sale'],
                              Record)
        self.assertEqual(
            self.client.plugins.module_utils.modules['sale']._object.name,
            'ir.module.module')

    def test_20_modules_dir(self):
        self.assertIn('m_sale', dir(self.client.plugins.module_utils))
        self.assertIn('modules', dir(self.client.plugins.module_utils))
        self.assertIn('update_module_list',
                      dir(self.client.plugins.module_utils))
        self.assertIn('installed_modules',
                      dir(self.client.plugins.module_utils))

    def test_25_module_getitem(self):
        res = self.client.plugins.module_utils['sale']
        self.assertIsInstance(res, Record)
        self.assertEqual(res._object.name, 'ir.module.module')

        from odoo_rpc_client.plugins.module_utils import ModuleObject

        self.assertIn(ModuleObject, res._object.__class__.__bases__)

    def test_30_module_getattr(self):
        res = self.client.plugins.module_utils.m_sale

        self.assertIsInstance(res, Record)
        self.assertEqual(res._object.name, 'ir.module.module')

        from odoo_rpc_client.plugins.module_utils import ModuleObject

        self.assertIn(ModuleObject, res._object.__class__.__bases__)

        with self.assertRaises(AttributeError):
            self.client.plugins.module_utils.m_unexistent_module

    def test_35_module_contains(self):
        self.assertIsNone(self.client.plugins.module_utils._modules)
        self.assertIn('sale', self.client.plugins.module_utils)
        self.assertNotIn(
            'some strange unexistent module',
            self.client.plugins.module_utils)
        self.assertIsNotNone(self.client.plugins.module_utils._modules)

    def test_40_module_install(self):
        smod = self.client.plugins.module_utils.m_sale

        if smod.state == 'installed':
            raise unittest.SkipTest('Module already installed')

        self.assertNotEqual(smod.state, 'installed')
        smod.install()
        smod.refresh()  # reread data from database
        self.assertEqual(smod.state, 'installed')

    def test_45_update_module_list(self):
        # touch modules proerty to load module info
        self.client.plugins.module_utils.modules

        self.assertIsNotNone(self.client.plugins.module_utils._modules)
        self.client.plugins.module_utils.update_module_list()
        self.assertIsNone(self.client.plugins.module_utils._modules)

    def test_50_module_upgrade(self):
        smod = self.client.plugins.module_utils.m_sale

        self.assertEqual(smod.state, 'installed')
        smod.upgrade()  # just call it

    def test_55_installed_modules(self):
        modules = self.client.plugins.module_utils.installed_modules
        modules2 = self.client['ir.module.module'].search_records(
            [('state', '=', 'installed')])
        self.assertItemsEqual(modules, modules2)


class Test_26_Plugin_ExternalIDS(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        self.data_obj = self.client['ir.model.data']
        self.partner_obj = self.client['res.partner']
        main_partner_domain = [
            ('module', '=', 'base'),
            ('model', '=', 'res.partner'),
            ('name', '=', 'main_partner'),
        ]
        self.main_partner_data = self.data_obj.search_records(
            main_partner_domain)[0]
        self.main_partner = self.partner_obj.browse(
            self.main_partner_data.res_id)

    def test_10_init_module_utils(self):
        self.assertNotIn('external_ids', self.client.plugins)
        self.assertIn('Test', self.client.plugins)

        import odoo_rpc_client.plugins.external_ids  # noqa

        self.assertIn('external_ids', self.client.plugins)

    def test_20_get_for__record(self):
        data = self.client.plugins.external_ids.get_for(self.main_partner)
        self.assertIsInstance(data, RecordList)
        self.assertEqual(data.object.name, 'ir.model.data')
        self.assertEqual(len(data), 1)

        data0 = data[0]
        self.assertEqual(data0.model, 'res.partner')
        self.assertEqual(data0.res_id, self.main_partner.id)
        self.assertEqual(data0.module, 'base')

    def test_21_get_for__recordlist(self):
        recordlist = self.client['res.partner'].search_records([], limit=10)
        res = self.client.plugins.external_ids.get_for(recordlist)

        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.object.name, 'ir.model.data')

        for data in res:
            self.assertIn(data.res_id, recordlist.ids)

    def test_22_get_for__tuple(self):
        data = self.client.plugins.external_ids.get_for(('res.partner',
                                                         self.main_partner.id))
        self.assertIsInstance(data, RecordList)
        self.assertEqual(data.object.name, 'ir.model.data')
        self.assertEqual(len(data), 1)

        data0 = data[0]
        self.assertEqual(data0.model, 'res.partner')
        self.assertEqual(data0.res_id, self.main_partner.id)
        self.assertEqual(data0.module, 'base')

    def test_23_get_for__str(self):
        data = self.client.plugins.external_ids.get_for('base.main_partner')
        self.assertIsInstance(data, RecordList)
        self.assertEqual(data.object.name, 'ir.model.data')
        self.assertEqual(len(data), 1)

        data0 = data[0]
        self.assertEqual(data0.model, 'res.partner')
        self.assertEqual(data0.res_id, self.main_partner.id)
        self.assertEqual(data0.module, 'base')

    def test_23_get_for__str_module(self):
        data = self.client.plugins.external_ids.get_for('main_partner', 'base')
        self.assertIsInstance(data, RecordList)
        self.assertEqual(data.object.name, 'ir.model.data')
        self.assertEqual(len(data), 1)

        data0 = data[0]
        self.assertEqual(data0.model, 'res.partner')
        self.assertEqual(data0.res_id, self.main_partner.id)
        self.assertEqual(data0.module, 'base')

    def test_24_get_for__bad_value(self):
        with self.assertRaises(ValueError):
            self.client.plugins.external_ids.get_for(None, 'base')

    def test_25_get_for__unexisting_xml_id(self):
        data = self.client.plugins.external_ids.get_for(
            'base.unexisting_xml_id')
        self.assertIsInstance(data, RecordList)
        self.assertEqual(data.length, 0)
        self.assertFalse(data)

    def test_26_get_for__wrong_xml_id(self):
        with self.assertRaises(ValueError):
            self.client.plugins.external_ids.get_for('bad_xml_id')

    def test_30_get_xmlid(self):
        xml_id = self.client.plugins.external_ids.get_xmlid(self.main_partner)
        self.assertEqual(xml_id, "base.main_partner")

        # Create new partner without xml_id
        new_partner_id = self.client[
            'res.partner'].create({'name': 'Test partner'})
        new_partner = self.client['res.partner'].browse(new_partner_id)

        no_xml_id = self.client.plugins.external_ids.get_xmlid(new_partner)
        self.assertFalse(no_xml_id)

        # Cleanup, remove created partner
        new_partner.unlink()

    def test_35_get_record(self):
        mpartner = self.client.plugins.external_ids.get_record(
            'base.main_partner')
        self.assertEqual(mpartner, self.main_partner)

        no_partner = self.client.plugins.external_ids.get_record(
            'base.unexisting_xml_id')
        self.assertFalse(no_partner)

    def test_40_record_as_xmlid(self):
        xml_id = self.main_partner.as_xmlid()
        self.assertEqual(xml_id, "base.main_partner")

        new_partner_id = self.client[
            'res.partner'].create({'name': 'Test partner'})
        new_partner = self.client['res.partner'].browse(new_partner_id)

        no_xml_id = new_partner.as_xmlid()
        self.assertFalse(no_xml_id)

        # Cleanup, remove created partner
        new_partner.unlink()



import six
import numbers
import collections

from . import (BaseTestCase,
               mock)
from ..client import Client
from ..orm.record import (Record,
                          RecordList,
                          get_record_list)
from ..orm.cache import (empty_cache,
                         ObjectCache,
                         Cache)
from ..orm.object import Object
from ..exceptions import ConnectorError


class Test_20_Object(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        self.object = self.client.get_obj('res.partner')

    def test_dir(self):
        self.assertIn('read', dir(self.object))
        self.assertIn('search', dir(self.object))
        self.assertIn('write', dir(self.object))
        self.assertIn('unlink', dir(self.object))
        self.assertIn('create', dir(self.object))

        # test if normal mehtods avalilable in dir(object)
        self.assertIn('search_records', dir(self.object))
        self.assertIn('browse', dir(self.object))

    def test_getttr(self):
        self.assertEqual(self.object.search.__name__, 'search')

        self.assertEqual(self.object.some_partner_method.__name__,
                         '%s:some_partner_method' % self.object.name)
        self.assertTrue(self.object.some_partner_method.__x_stdcall__)

        # Test that attibute error is raised on access on private methods
        with self.assertRaises(AttributeError):
            self.object._do_smthing_private

    def test_call_unexistent_method(self):
        # method wrapper will be created
        self.assertEqual(self.object.some_unexisting_mehtod.__name__,
                         'res.partner:some_unexisting_mehtod')

        # but exception should be raised
        with self.assertRaises(ConnectorError):
            self.object.some_unexisting_mehtod([1])

    def test_model(self):
        self.assertIsInstance(self.object.model, Record)
        self.assertEqual(self.object.name, self.object.model.model)
        self.assertEqual(self.object.model, self.object._model)

        # this will check that model_name is result of name_get on model record
        self.assertEqual(self.object.model_name, self.object.model._name)

    def test_search(self):
        res = self.object.search([('id', '=', 1)])
        self.assertIsInstance(res, list)
        self.assertEqual(res, [1])

        res = self.object.search([('id', '=', 1)], count=1)
        self.assertIsInstance(res, numbers.Integral)
        self.assertEqual(res, 1)

    def test_search_read(self):
        res = self.object.search_read([('id', '=', 1)], fields=['name'])
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0], dict)
        self.assertIn('id', res[0])
        self.assertIn('name', res[0])
        self.assertEqual(res[0]['id'], 1)

    def test_search_records(self):
        res = self.object.search_records([('id', '=', 1)])
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.length, 1)
        self.assertEqual(res[0].id, 1)

        res = self.object.search_records([('id', '=', 99999)])
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.length, 0)

        res = self.object.search_records([('id', '=', 1)], count=1)
        self.assertIsInstance(res, numbers.Integral)
        self.assertEqual(res, 1)

        # test search_records with read_fields argument
        res = self.object.search_records([],
                                         read_fields=['name', 'country_id'],
                                         limit=10)
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.length, 10)
        self.assertEqual(len(res._lcache), res.length)
        for record in res:
            self.assertEqual(len(record._data), 3)
            self.assertItemsEqual(list(record._data),
                                  ['id', 'name', 'country_id'])

    def test_read_records(self):
        # read one record
        res = self.object.read_records(1)
        self.assertIsInstance(res, Record)
        self.assertEqual(res.id, 1)

        # read set of records
        res = self.object.read_records([1])
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.length, 1)
        self.assertEqual(res[0].id, 1)

        # try to call read_records with bad argument
        with self.assertRaises(ValueError):
            self.object.read_records(None)

        # Test read with specified fields
        record = self.object.read_records(1,
                                          ['name', 'country_id'])
        self.assertEqual(len(record._data), 3)
        self.assertItemsEqual(list(record._data),
                              ['id', 'name', 'country_id'])

    def test_browse(self):
        with mock.patch.object(self.object,
                               'read_records') as fake_read_records:
            self.object.browse(1)
            fake_read_records.assert_called_with(1)

        with mock.patch.object(self.object,
                               'read_records') as fake_read_records:
            self.object.browse([1])
            fake_read_records.assert_called_with([1])

        with mock.patch.object(self.object,
                               'read_records') as fake_read_records:
            self.object.browse(None)
            fake_read_records.assert_called_with(None)

    def test_object_equal(self):
        self.assertEqual(self.object, self.client['res.partner'])
        self.assertIs(self.object, self.client['res.partner'])
        self.assertNotEqual(self.object, self.client['res.users'])
        self.assertIsNot(self.object, self.client['res.users'])

    def test_str(self):
        self.assertEqual(str(self.object), u"Object ('res.partner')")

    def test_repr(self):
        self.assertEqual(repr(self.object), str(self.object))

    def test_object_specific_extension(self):
        class MyProductObject(Object):

            class Meta:
                name = 'res.users'

            def test_previously_unexistent_method(self):
                return "Method Ok"

        # reload client caches
        self.client.clean_caches()

        # Newly defined object method must be present in res.users model
        res = self.client['res.users'].test_previously_unexistent_method()
        self.assertEqual(res, "Method Ok")

        # but must not be present in other objects / models
        with self.assertRaises(ConnectorError):
            self.client['res.partner'].test_previously_unexistent_method()

    def test_create_write_unlink(self):
        new_partner_id = self.object.create({'name': 'New Partner'})

        self.assertIsInstance(new_partner_id, six.integer_types)

        self.object.write([new_partner_id], {'name': 'New Partner Name'})

        new_name = self.object.read(new_partner_id, ['name'])['name']

        self.assertEqual(new_name, 'New Partner Name')

        self.assertEqual(self.object.search([('id', '=', new_partner_id)],
                                            count=True),
                         1)
        self.object.unlink([new_partner_id])
        self.assertEqual(self.object.search([('id', '=', new_partner_id)],
                                            count=True),
                         0)


class Test_21_Record(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        self.object = self.client.get_obj('res.partner')
        self.record = self.object.browse(1)

    def test_dir(self):
        self.assertIn('read', dir(self.record))
        self.assertIn('write', dir(self.record))
        self.assertIn('unlink', dir(self.record))

        # test if normal mehtods avalilable in dir(object)
        self.assertIn('refresh', dir(self.record))

    def test_client_property(self):
        self.assertIs(self.record._client, self.client)
        self.assertIs(self.record._object.client, self.client)
        self.assertIs(self.object.client, self.client)

    def test_as_dict(self):
        rdict = self.record.as_dict

        self.assertIsInstance(rdict, dict)
        self.assertIsNot(rdict, self.record._data)
        self.assertItemsEqual(rdict, self.record._data)

        # test that changes to rdict will not calue changes to record's data
        rdict['new_key'] = 'new value'

        self.assertIn('new_key', rdict)
        self.assertNotIn('new_key', self.record._data)

    def test_str(self):
        self.assertEqual(str(self.record),
                         u"R(res.partner, 1)[%s]"
                         u"" % (self.record.name_get()[0][1]))

    def test_repr(self):
        self.assertEqual(str(self.record), repr(self.record))

    def test_name_get(self):
        self.assertEqual(self.record._name, self.record.name_get()[0][1])

    def test_record_equal(self):
        rec1 = self.record

        rec_list = self.object.search_records([('id', '=', 1)])
        self.assertIsInstance(rec_list, RecordList)
        self.assertEqual(rec_list.length, 1)

        rec2 = rec_list[0]
        self.assertEqual(rec1, rec2)
        self.assertEqual(rec1.id, rec2.id)
        self.assertEqual(rec1._name, rec2._name)

        # Test that equality with simple integers works
        self.assertEqual(rec1, rec2.id)
        self.assertEqual(rec1.id, rec2)

        self.assertNotEqual(rec1, None)
        self.assertNotEqual(rec1, 2)

    def test_getitem(self):
        self.assertEqual(self.record['id'], self.record.id)
        with self.assertRaises(KeyError):
            self.record['some_unexistent_field']

    def test_getattr(self):
        # Check that, if we try to get unexistent field, result will be method
        # wrapper for object method
        f = self.record.some_unexistent_field
        self.assertTrue(callable(f))

    def test_record_to_int(self):
        self.assertIs(int(self.record), 1)

    def test_record_hash(self):
        self.assertEqual(hash(self.record),
                         hash((self.record._object.name, self.record.id)))

    def test_record_relational_fields(self):
        # read data from res.partner:child_ids field
        res = self.record.child_ids

        self.assertIsInstance(res, RecordList)
        self.assertGreaterEqual(res.length, 1)
        self.assertIsInstance(res[0], Record)
        self.assertEqual(res[0]._object.name, 'res.partner')

        # test many2one
        self.assertIsInstance(res[0].parent_id, Record)
        self.assertIsNot(res[0].parent_id, self.record)
        self.assertEqual(res[0].parent_id, self.record)

        # test that empty many2one field is avaluated as False
        self.assertIs(self.record.user_id, False)

        # test that empty x2many field is evaluated as empty RecordList
        self.assertIsInstance(self.record.user_ids, RecordList)
        self.assertEqual(self.record.user_ids.length, 0)

    def test_record_refresh(self):
        # read all data for record
        self.record.read()

        # read company_id field
        self.record.company_id.name

        # check that data had been loaded
        self.assertGreater(len(self.record._data.keys()), 5)

        # test before refresh
        self.assertEqual(len(self.record._cache.keys()), 2)
        self.assertIn('res.partner', self.record._cache)
        self.assertIn('res.company', self.record._cache)
        self.assertIn(
            len(list(self.record._cache['res.company'].values())[0]), [2, 3])
        self.assertIn(
            'name',
            list(
                self.record._cache['res.company'].values())[0])

        # refresh record
        self.record.refresh()

        # test after refresh
        self.assertEqual(len(self.record._data.keys()), 1)
        self.assertItemsEqual(list(self.record._data), ['id'])
        self.assertEqual(len(self.record._cache.keys()), 2)
        self.assertIn('res.partner', self.record._cache)
        self.assertIn('res.company', self.record._cache)
        self.assertEqual(
            len(list(self.record._cache['res.company'].values())[0]), 1)
        self.assertNotIn(
            'name',
            list(
                self.record._cache['res.company'].values())[0])

    def test_record_specific_extension(self):
        class MyProductRecord(Record):

            class Meta:
                object_name = 'res.users'

            def test_previously_unexistent_record_method(self):
                return "Method Ok Record %s" % self.id

        # Product records must have this method
        user = self.client['res.users'].search_records([], limit=1)[0]
        res = user.test_previously_unexistent_record_method()
        self.assertEqual(res, "Method Ok Record %s" % user.id)

        # other records must not have it
        r = self.client['res.partner'].search_records([], limit=1)[0]
        with self.assertRaises(ConnectorError):
            r.test_previously_unexistent_record_method()

    def test_copy(self):
        res = self.record.copy()

        self.assertIsInstance(res, Record)
        self.assertNotEqual(res.id, self.record.id)

        res = self.record.copy(default={'ref': 'My Test Copy Partner Ref'})
        self.assertIsInstance(res, Record)
        self.assertNotEqual(res.id, self.record.id)
        self.assertNotEqual(res.ref, self.record.ref)
        self.assertEqual(res.ref, 'My Test Copy Partner Ref')

    def test_get_existing_field(self):
        rec = self.record
        self.assertEqual(rec.get('name'), self.record.name)
        self.assertEqual(rec.get('name', 'default'), self.record.name)

    def test_get_unexisting_field(self):
        rec = self.record
        with self.assertRaises(KeyError):
            rec.get('unexisting_name')

        self.assertEqual(rec.get('unexisting_name', 'default'), 'default')


class Test_22_RecordList(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        self.object = self.client.get_obj('res.partner')
        self.all_obj_ids = self.object.search([])
        self.obj_ids = self.object.search([],
                                          limit=len(self.all_obj_ids) - 2)
        self.recordlist = self.object.read_records(self.obj_ids)

    def test_dir(self):
        self.assertIn('read', dir(self.recordlist))
        self.assertIn('write', dir(self.recordlist))
        self.assertIn('unlink', dir(self.recordlist))

        # test if normal mehtods avalilable in dir(object)
        self.assertIn('refresh', dir(self.recordlist))

    def test_ids(self):
        self.assertSequenceEqual(self.recordlist.ids, self.obj_ids)

    def test_length(self):
        self.assertEqual(self.recordlist.length, len(self.obj_ids))
        self.assertEqual(len(self.recordlist), len(self.obj_ids))

    def test_recods(self):
        self.assertIsInstance(self.recordlist.records, list)
        self.assertIsInstance(self.recordlist.records[0], Record)

    def test_str(self):
        self.assertEqual(
            str(self.recordlist),
            u"RecordList(res.partner): length=%d" % len(self.obj_ids))

    def test_repr(self):
        self.assertEqual(repr(self.recordlist), str(self.recordlist))

    def test_getitem(self):
        id1 = self.obj_ids[0]
        id2 = self.obj_ids[-1]

        id_slice = self.obj_ids[2:15:2]

        self.assertIsInstance(self.recordlist[0], Record)
        self.assertEqual(self.recordlist[0].id, id1)

        self.assertIsInstance(self.recordlist[-1], Record)
        self.assertEqual(self.recordlist[-1].id, id2)

        res = self.recordlist[2:15:2]
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.length, len(id_slice))
        self.assertSequenceEqual(res.ids, id_slice)

        with self.assertRaises(IndexError):
            self.recordlist[100]

    def test_getattr(self):
        with mock.patch.object(self.object,
                               'some_server_method') as fake_method:
            # bug override in mock object (python 2.7)
            if not getattr(fake_method, '__name__', False):
                fake_method.__name__ = fake_method.name
            self.recordlist.some_server_method('arg1', 'arg2')
            fake_method.assert_called_with(self.recordlist.ids, 'arg1', 'arg2')

    def test_delitem(self):
        r = self.recordlist[5]
        self.assertEqual(len(self.recordlist), len(self.obj_ids))

        del self.recordlist[5]

        self.assertEqual(len(self.recordlist), len(self.obj_ids) - 1)
        self.assertNotIn(r, self.recordlist)

    def test_setitem(self):
        rec = self.object.search_records(
            [('id', 'not in', self.recordlist.ids)], limit=1)[0]

        old_rec = self.recordlist[8]

        self.assertEqual(len(self.recordlist), len(self.obj_ids))
        self.assertNotIn(rec, self.recordlist)
        self.assertIn(old_rec, self.recordlist)

        self.recordlist[8] = rec

        self.assertEqual(len(self.recordlist), len(self.obj_ids))
        self.assertIn(rec, self.recordlist)
        self.assertNotIn(old_rec, self.recordlist)

        with self.assertRaises(ValueError):
            self.recordlist[5] = 25

    def test_contains(self):
        rid = self.obj_ids[0]
        rec = self.object.read_records(rid)

        brid = self.object.search([('id', 'not in', self.obj_ids)], limit=1)[0]
        brec = self.object.read_records(brid)

        self.assertIn(rid, self.recordlist)
        self.assertIn(rec, self.recordlist)

        self.assertNotIn(brid, self.recordlist)
        self.assertNotIn(brec, self.recordlist)

        self.assertNotIn(None, self.recordlist)

    def test_insert_record(self):
        rec = self.object.search_records(
            [('id', 'not in', self.recordlist.ids)], limit=1)[0]

        self.assertEqual(len(self.recordlist), len(self.obj_ids))
        self.assertNotIn(rec, self.recordlist)

        self.recordlist.insert(1, rec)

        self.assertEqual(len(self.recordlist), len(self.obj_ids) + 1)
        self.assertIn(rec, self.recordlist)
        self.assertEqual(self.recordlist[1], rec)

    def test_insert_by_id(self):
        rec = self.object.search_records(
            [('id', 'not in', self.recordlist.ids)], limit=1)[0]

        self.assertEqual(len(self.recordlist), len(self.obj_ids))
        self.assertNotIn(rec, self.recordlist)

        self.recordlist.insert(1, rec.id)

        self.assertEqual(len(self.recordlist), len(self.obj_ids) + 1)
        self.assertIn(rec, self.recordlist)
        self.assertEqual(self.recordlist[1], rec)

    def test_insert_bad_value(self):
        rec = self.object.search_records(
            [('id', 'not in', self.recordlist.ids)], limit=1)[0]

        self.assertEqual(len(self.recordlist), len(self.obj_ids))
        self.assertNotIn(rec, self.recordlist)

        with self.assertRaises(AssertionError):
            self.recordlist.insert(1, "some strange type")

    def test_mapped_1_simple_field(self):
        res = self.recordlist.mapped('name')
        self.assertIsInstance(res, list)
        self.assertIsInstance(res[0], six.string_types)
        self.assertEqual(len(res), len(set(p.name for p in self.recordlist)))

    def test_mapped_2_m2o_field(self):
        res = self.recordlist.mapped('parent_id')
        self.assertIsInstance(res, RecordList)
        self.assertIsInstance(res[0], Record)

        # TODO: rewrite this to test that items was uniquifyed
        self.assertEqual(
            len(res),
            len(set([r.parent_id for r in self.recordlist if r.parent_id])))

    def test_mapped_3_m2o_dot_char_field(self):
        res = self.recordlist.mapped('parent_id.name')
        self.assertIsInstance(res, list)
        self.assertGreaterEqual(len(res), 1)
        self.assertIsInstance(res[0], six.string_types)

        # TODO: rewrite this to test that items was uniquifyed
        self.assertEqual(
            len(res),
            len(set([r.parent_id.name
                     for r in self.recordlist if r.parent_id])))

    def test_mapped_4_m2o_dot_char_field(self):
        res = self.recordlist.mapped('country_id.code')
        self.assertIsInstance(res, list)
        self.assertGreaterEqual(len(res), 1)
        self.assertIsInstance(res[0], six.string_types)

        # TODO: rewrite this to test that items was uniquifyed
        self.assertEqual(
            len(res),
            len(set([r.country_id.code
                     for r in self.recordlist if r.country_id])))

    def test_mapped_5_m2o_dot_o2m_field(self):
        res = self.recordlist.mapped('user_ids')
        self.assertIsInstance(res, RecordList)
        self.assertGreaterEqual(len(res), 1)
        self.assertIsInstance(res[0], Record)
        self.assertEqual(res[0]._object.name, 'res.users')
        # TODO: implement some additional checks

    def test_prefetch(self):
        cache = self.recordlist._cache
        lcache = self.recordlist._lcache

        # check that cache is only filled with ids
        self.assertEqual(len(lcache), self.recordlist.length)
        for record in self.recordlist:
            # Note that record._data is a property, which means
            # record._lcache[record.id]. _data property is dictionary.
            self.assertEqual(len(record._data), 1)
            self.assertItemsEqual(list(record._data), ['id'])

        # prefetch normal field
        self.recordlist.prefetch('name')

        self.assertEqual(len(self.recordlist._lcache), self.recordlist.length)
        for record in self.recordlist:
            self.assertEqual(len(record._data), 2)
            self.assertItemsEqual(list(record._data), ['id', 'name'])

        # check that cache contains only res.partner object cache
        self.assertEqual(len(cache), 1)
        self.assertIn('res.partner', cache)
        self.assertNotIn('res.country', cache)

        # prefetch related field name of caountry and country code
        self.recordlist.prefetch('country_id.name', 'country_id.code')

        # test that cache now contains two objects ('res.partner',
        # 'res.country')
        self.assertEqual(len(cache), 2)
        self.assertIn('res.partner', cache)
        self.assertIn('res.country', cache)

        c_cache = cache['res.country']
        # if in some cases selected partners have no related countries, raise
        # error
        country_checked = False
        for record in self.recordlist:
            # test that country_id field was added to partner's cache
            self.assertEqual(len(record._data), 3)
            self.assertItemsEqual(
                list(record._data),
                ['id', 'name', 'country_id'])

            # if current partner have related country_id
            #
            # Note, here check 'country_id' via '_data' property to avoid lazy
            # loading of data.
            country_id = record._data['country_id']

            # if data is in form [id, <name_get result>]
            if isinstance(country_id, collections.Iterable):
                country_id = country_id[0]
                country_is_list = True

            if country_id:
                country_checked = True

                # test, that there are some data for this country_id in country
                # cache
                self.assertIn(country_id, c_cache)

                # Note that, in case, of related many2one fields, Odoo may
                # return list, with Id and resutlt of name_get method.
                # thus, we program will imediatly cache this value
                if country_is_list:
                    self.assertEqual(len(c_cache[country_id]), 4)
                    self.assertItemsEqual(
                        list(c_cache[country_id]),
                        ['id', 'name', 'code', '__name_get_result'])
                else:
                    self.assertEqual(len(c_cache[country_id]), 4)
                    self.assertItemsEqual(
                        list(c_cache[country_id]),
                        ['id', 'name', 'code', '__name_get_result'])

        self.assertTrue(
            country_checked,
            "Country must be checked. "
            "may be there are wrong data in test database")

    def test_copy(self):
        # test that cipied list and original list have same cache instance,
        # when new_cache arg is set to default value False
        clist = self.recordlist.copy()
        self.assertIs(clist._cache, self.recordlist._cache)

        # When new_cache arg set to True, then new_automaticaly generated cache
        # will be used
        clist = self.recordlist.copy(new_cache=True)
        self.assertIsNot(clist._cache, self.recordlist._cache)

        # test when Cache instance passed to new_cache argument
        cache = Cache(self.recordlist.object.client)
        clist = self.recordlist.copy(new_cache=cache)
        self.assertIs(clist._cache, cache)

        # Test that with wrong argument copy will raise
        with self.assertRaises(ValueError):
            self.recordlist.copy(new_cache=25)

    def test_sorting(self):
        def to_names(rlist):
            """ Convert recordlist to list of names
            """
            return [r.name for r in rlist]

        names = to_names(self.recordlist)

        self.assertSequenceEqual(sorted(names),
                                 to_names(sorted(self.recordlist,
                                                 key=lambda x: x.name)))
        self.assertSequenceEqual(sorted(names, reverse=True),
                                 to_names(sorted(self.recordlist,
                                                 key=lambda x: x.name,
                                                 reverse=True)))
        self.assertSequenceEqual(list(reversed(names)),
                                 to_names(reversed(self.recordlist)))

        # test recordlist sort methods
        rlist = self.recordlist.copy()
        rnames = names[:]  # copy list
        rlist.sort(key=lambda x: x.name)   # inplace sort
        rnames.sort()  # inplace sort
        self.assertSequenceEqual(rnames, to_names(rlist))

        # test recordlist reverse method
        rlist = self.recordlist.copy()
        rnames = names[:]  # copy list
        rlist.reverse()    # inplace reverse
        rnames.reverse()   # inplace reverse
        self.assertSequenceEqual(rnames, to_names(rlist))

    def test_search(self):
        # TODO: test for context
        with mock.patch.object(self.object, 'search') as fake_method:
            self.recordlist.search([('id', '!=', 1)], limit=5)
            fake_method.assert_called_with(
                [('id', 'in', self.recordlist.ids), ('id', '!=', 1)], limit=5)

    def test_search_records(self):
        # TODO: test for context
        with mock.patch.object(self.object, 'search_records') as fake_method:
            self.recordlist.search_records([('id', '!=', 1)], limit=4)
            fake_method.assert_called_with(
                [('id', 'in', self.recordlist.ids), ('id', '!=', 1)], limit=4)

    def test_read(self):
        # TODO: test for context
        #       or remove this test and method, because getattr pass context
        #       too
        with mock.patch.object(self.object, 'read') as fake_method:
            self.recordlist.read(['name'])
            fake_method.assert_called_with(self.recordlist.ids, ['name'])

    def test_filter(self):
        res = self.recordlist.filter(lambda x: x.id % 2 == 0)
        expected_ids = [r.id for r in self.recordlist if r.id % 2 == 0]
        self.assertIsInstance(res, RecordList)
        self.assertEqual(res.ids, expected_ids)

    def test_group_by(self):
        res = self.recordlist.group_by(lambda x: x.id % 2 == 0)
        self.assertIsInstance(res, collections.defaultdict)
        self.assertItemsEqual(res.keys(), [True, False])
        # TODO: write better test

        res = self.recordlist.group_by('country_id')
        self.assertIsInstance(res, collections.defaultdict)

    def test_existing(self):
        # all existing object ids
        all_obj_ids = self.object.search([], limit=False)

        # generate 10 unexisting ids
        unexistent_ids = list(range(max(all_obj_ids) + 1,
                                    max(all_obj_ids) + 40,
                                    4))
        self.assertEqual(len(unexistent_ids), 10)

        # test simple existense
        rlist = get_record_list(self.object, all_obj_ids[:10] + unexistent_ids)
        self.assertEqual(len(rlist), 20)
        elist = rlist.existing()
        self.assertEqual(len(elist), 10)
        self.assertItemsEqual(elist.ids, all_obj_ids[:10])

        # test existense with repeated items
        rlist_ids = all_obj_ids[:10] + unexistent_ids + all_obj_ids[:5]
        rlist = get_record_list(self.object, rlist_ids)
        self.assertEqual(len(rlist), 25)

        # with uniqify=True (defualt)
        elist = rlist.existing()
        self.assertEqual(len(elist), 10)
        self.assertItemsEqual(elist.ids, all_obj_ids[:10])

        # with uniqify=False
        elist = rlist.existing(uniqify=False)
        self.assertEqual(len(elist), 15)
        self.assertItemsEqual(elist.ids, all_obj_ids[:10] + all_obj_ids[:5])

    def test_refresh(self):
        # save cache pointers to local namespase to simplify access to it
        cache = self.recordlist._cache
        pcache = cache['res.partner']  # partner cache
        ccache = cache['res.country']  # country cache

        # load data to record list
        self.recordlist.prefetch('name', 'country_id.name', 'country_id.code')

        # create related records. This is still required, becuase prefetch
        # just fills cache without creating record instances
        for rec in self.recordlist:
            rec.country_id

        self.assertEqual(len(pcache), len(self.recordlist))
        self.assertGreaterEqual(
            len(ccache),
            1)  # there are atleast on country in cache

        clen = len(ccache)

        for data in pcache.values():
            self.assertItemsEqual(list(data), ['id', 'name', 'country_id'])

        for data in ccache.values():
            if '__name_get_result' in data:
                self.assertItemsEqual(
                    list(data), [
                        'id', 'name', 'code', '__name_get_result'])
            else:
                self.assertItemsEqual(list(data), ['id', 'name', 'code'])

        # refresh recordlist
        self.recordlist.refresh()

        self.assertEqual(len(pcache), len(self.recordlist))
        self.assertEqual(len(ccache), clen)

        # test that data for each partner is empty
        for data in pcache.values():
            self.assertItemsEqual(list(data), ['id'])

        # test that data for each country is empty
        for data in ccache.values():
            self.assertItemsEqual(list(data), ['id'])


class Test_23_Cache(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        self.cache = empty_cache(self.client)

    def test_client(self):
        self.assertIs(self.cache.client, self.client)

    def test_missing(self):
        obj_cache = self.cache['res.partner']
        self.assertIsInstance(obj_cache, ObjectCache)

        # Test that KeyError is raised on attempt to get cache for object that
        # does not exist in client database
        with self.assertRaises(KeyError):
            self.cache['unexisting.object']

    def test_missing_local(self):
        obj_cache = self.cache['res.partner']

        self.assertFalse(bool(obj_cache))
        self.assertNotIn(42, obj_cache)
        res = obj_cache[42]
        self.assertIsInstance(res, dict)
        self.assertIn('id', res)
        self.assertEqual(len(res), 1)
        self.assertIs(res['id'], 42)
        self.assertIn(42, obj_cache)

    def test_update_keys(self):
        obj_cache = self.cache['res.partner']

        self.assertFalse(bool(obj_cache))
        self.assertNotIn(1, obj_cache)
        self.assertNotIn(2, obj_cache)
        self.assertNotIn(3, obj_cache)
        self.assertNotIn(4, obj_cache)

        # update cache with keys
        obj_cache.update_keys([1, 2, 3])

        self.assertIn(1, obj_cache)
        self.assertEqual(len(obj_cache[1]), 1)
        self.assertIn('id', obj_cache[1])
        self.assertIs(obj_cache[1]['id'], 1)

        self.assertIn(2, obj_cache)
        self.assertEqual(len(obj_cache[2]), 1)
        self.assertIn('id', obj_cache[2])
        self.assertIs(obj_cache[2]['id'], 2)

        self.assertIn(3, obj_cache)
        self.assertEqual(len(obj_cache[3]), 1)
        self.assertIn('id', obj_cache[3])
        self.assertIs(obj_cache[3]['id'], 3)

        self.assertNotIn(4, obj_cache)

        # add new cache keys
        obj_cache.update_keys([3, 4, 6])

        self.assertIn(3, obj_cache)
        self.assertEqual(len(obj_cache[3]), 1)
        self.assertIn('id', obj_cache[3])
        self.assertIs(obj_cache[3]['id'], 3)

        self.assertIn(4, obj_cache)
        self.assertEqual(len(obj_cache[4]), 1)
        self.assertIn('id', obj_cache[4])
        self.assertIs(obj_cache[4]['id'], 4)

        self.assertNotIn(5, obj_cache)

        self.assertIn(6, obj_cache)
        self.assertEqual(len(obj_cache[6]), 1)
        self.assertIn('id', obj_cache[6])
        self.assertIs(obj_cache[6]['id'], 6)

    def test_get_ids_to_read(self):
        obj_cache = self.cache['res.partner']

        # update cache with keys
        obj_cache.update_keys([1, 2, 3, 4, 5])

        self.assertItemsEqual(obj_cache.get_ids_to_read('name'),
                              [1, 2, 3, 4, 5])

        # fill field 'name' in first two cache items
        obj_cache[1]['name'] = 'item 1'
        obj_cache[2]['name'] = 'item 2'

        # See that first two item are not mentioned in result, because they
        # already have value for field 'name'
        self.assertItemsEqual(obj_cache.get_ids_to_read('name'), [3, 4, 5])

        # See that if we get ids to read for fields 'name' and 'address', than
        # all ids will be returned again. this is because no cache item have
        # field 'aaddress' filled
        self.assertItemsEqual(obj_cache.get_ids_to_read('name', 'address'),
                              [1, 2, 3, 4, 5])

        # test that if we change order of fields, nothing will change
        self.assertItemsEqual(obj_cache.get_ids_to_read('address', 'name'),
                              [1, 2, 3, 4, 5])

        # Add value for field 'address' to cache items #3 and #2
        obj_cache[3]['address'] = 'Kyiv, Ukraine'
        obj_cache[2]['address'] = 'Zhytomyr, Ukraine'

        # ad test what ids will be returned
        self.assertItemsEqual(obj_cache.get_ids_to_read('address'), [1, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('name'), [3, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('address', 'name'),
                              [1, 3, 4, 5])

        # Ok, let's add field 'city' to last cache item (#5)
        obj_cache[5]['city'] = 'Kyiv'

        # And look what wi have
        self.assertItemsEqual(obj_cache.get_ids_to_read('address'), [1, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('name'), [3, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('city'), [1, 2, 3, 4])
        self.assertItemsEqual(obj_cache.get_ids_to_read('address', 'name'),
                              [1, 3, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('address', 'city'),
                              [1, 2, 3, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('name', 'city'),
                              [1, 2, 3, 4, 5])
        self.assertItemsEqual(obj_cache.get_ids_to_read('name',
                                                        'address',
                                                        'city'),
                              [1, 2, 3, 4, 5])

    def test_cache_field_str(self):
        obj_cache = self.cache['res.partner']

        # cache is empty
        self.assertFalse(bool(obj_cache))

        obj_cache.cache_field(5, 'char', 'name', 'Test Name')

        self.assertIn(5, obj_cache)
        self.assertIn('id', obj_cache[5])
        self.assertIs(obj_cache[5]['id'], 5)
        self.assertIn('name', obj_cache[5])
        self.assertEqual(obj_cache[5]['name'], 'Test Name')

    def test_cache_field_m2o_int(self):
        obj_cache = self.cache['res.partner']

        # cache is empty
        self.assertFalse(bool(obj_cache))

        obj_cache.cache_field(5, 'many2one', 'user_id', 7)

        # Test that all required data is present in obj_cache
        self.assertIn(5, obj_cache)
        self.assertIn('id', obj_cache[5])
        self.assertIs(obj_cache[5]['id'], 5)
        self.assertIn('user_id', obj_cache[5])
        self.assertEqual(obj_cache[5]['user_id'], 7)

        # Test that related cache filled
        self.assertIn('res.users', self.cache)
        rel_cache = self.cache['res.users']
        self.assertIsInstance(rel_cache, ObjectCache)
        self.assertIn(7, rel_cache)
        self.assertIn('id', rel_cache[7])
        self.assertIs(rel_cache[7]['id'], 7)

    def test_cache_field_m2o_tuple(self):
        obj_cache = self.cache['res.partner']

        # cache is empty
        self.assertFalse(bool(obj_cache))

        obj_cache.cache_field(5, 'many2one', 'user_id', (7, 'Super Admin'))

        # Test that all required data is present in obj_cache
        self.assertIn(5, obj_cache)
        self.assertIn('id', obj_cache[5])
        self.assertIs(obj_cache[5]['id'], 5)
        self.assertIn('user_id', obj_cache[5])

        # value is stored in cache as returned by server, without changed
        self.assertEqual(obj_cache[5]['user_id'], (7, 'Super Admin'))

        # Test that related cache filled
        self.assertIn('res.users', self.cache)
        rel_cache = self.cache['res.users']
        self.assertIsInstance(rel_cache, ObjectCache)
        self.assertIn(7, rel_cache)
        self.assertIn('id', rel_cache[7])
        self.assertIs(rel_cache[7]['id'], 7)

        # Test that name is saved as '__name_get_result' in related cache
        self.assertIn('__name_get_result', rel_cache[7])
        self.assertEqual(rel_cache[7]['__name_get_result'], 'Super Admin')

        # TODO: add tests for many2many and one2many field types

    def test_cache_update_context(self):
        cache = self.cache['res.partner']  # get object cache
        self.assertFalse(bool(cache.context))
        self.assertIs(cache.context, None)

        d = {'a': 5}
        ctx = cache.update_context(d)
        self.assertIsInstance(d, dict)
        self.assertDictEqual(ctx, d)
        self.assertIsNot(ctx, d)

        d.update({'b': 42})

        self.assertNotEqual(ctx, d)
        self.assertNotIn('b', ctx)
        self.assertIs(ctx, cache.context)

        ctx2 = cache.update_context({'c': 78})
        self.assertIs(cache.context, ctx2)
        self.assertIs(ctx2, ctx)
        self.assertIn('c', cache.context)
        self.assertIs(cache.context['c'], 78)
        self.assertIn('a', cache.context)
        self.assertNotIn('b', cache.context)



"""
"""

from ..plugin import Plugin
from ..orm.record import (Record,
                          RecordList)

import six


class ExternalIDS(Plugin):
    """ This plugin adds aditional methods to work with
        external_ids (xml_id) for Odoo records.
    """
    class Meta:
        name = "external_ids"

    def get_for(self, val, module=None):
        """ Return RecordList of 'ir.model.data' for val or False

            :param val: value to get 'ir.model.data' records for
            :param str module: module name to search 'ir.model.data' for
            :rtype: RecordList
            :return: RecordList with 'ir.model.data' records found
            :raises ValueError: if *val* argument could not be parsed

            ``val`` could be one of folowing types:

            - ``Record`` instance
            - ``RecordList`` instance
            - ``tuple(model, res_id)``, for example ``('res.partner', 5)``
            - ``str``, string in format 'module.name'.

            **Note**, in case of *val* is *str*: if *module*
            specified as parameter, then *val* supposed to be *name* only.
            For example, folowing calls are equal::

                cl.plugins.external_ids.get_for('base.group_configuration')
                cl.plugins.external_ids.get_for('group_configuration',
                                                module='base')
        """
        data_obj = self.client['ir.model.data']
        domain = [] if module is None else [('module', '=', module)]
        if isinstance(val, Record):
            domain += [('model', '=', val._object.name),
                       ('res_id', '=', val.id)]
        elif isinstance(val, RecordList):
            domain += [('model', '=', val._object.name),
                       ('res_id', 'in', val.ids)]
        elif isinstance(val, (tuple, list)) and len(val) == 2:
            model, res_id = val
            domain += [('model', '=', model),
                       ('res_id', '=', res_id)]
        elif isinstance(val, six.string_types):
            if module is None:
                try:
                    module, name = val.split('.')
                except ValueError:
                    raise ValueError(
                        "Bad xml_id passed. cannot fetch module name.")
            else:
                name = val

            domain = [('module', '=', module),
                      ('name', '=', name)]
        else:
            raise ValueError("Cannot parse value [%r]" % val)

        return data_obj.search_records(domain)

    def get_xmlid(self, val, module=None):
        """ Return *xml_id* for *val*.
            Note, that only first xml_id will be returned!

            :param val: look in documentation for
                `get_for
                <#odoo_rpc_client.plugins.external_ids.ExternalIDS.get_for>`__
                method
            :param str module: module name to search xml_id for
            :rtype: str
            :return: xml_id for *val* or False if not found
            :raises ValueError: if *val* argument could not be parsed

            Note, that if *module* specified as parametr, then *val*
            supposed to be *name* only
        """
        e_record = self.get_for(val, module=module)
        if e_record:
            return e_record[0].complete_name
        return False

    def get_record(self, xml_id, module=None):
        """ Return *Record* instance for specified xml_id

            :param str xml_id: string with xml_id to search record for
            :param str module: module name to search Record in
            :rtype: Record
            :return: Record for *val* or False if not found
            :raises ValueError: if *xml_id* argument could not be parsed
        """
        assert isinstance(xml_id, six.string_types), "xml_id must be string"
        e_record = self.get_for(xml_id, module=module)
        if e_record:
            e_record = e_record[0]
            return self.client[e_record.model].browse(e_record.res_id)
        return False


class Record__XMLIDS(Record):
    """ Simple class to add ability to get xmlid from record itself
    """
    def as_xmlid(self, module=None):
        """ Get xmlid for record

            :param str module: module to search xmlid in
            :return: xmlid for this record or False
            :rtype: str
        """
        return self._client.plugins.external_ids.get_xmlid(self, module=module)



import six
import collections

__all__ = ('empty_cache', 'Cache', 'ObjectCache')


class ObjectCache(dict):
    """ Cache for object / model data

        Automatically generates empty data dicts for records requested.
        Also contains object context
    """
    __slots__ = ('_root_cache', '_object', '_context')

    def __init__(self, root, obj, *args, **kwargs):
        self._root_cache = root
        self._object = obj
        self._context = kwargs.pop('context', None)
        super(ObjectCache, self).__init__(*args, **kwargs)

    @property
    def context(self):
        """ Return context instance related to this cache
        """
        return self._context

    def __missing__(self, key):
        self[key] = {'id': key}
        return self[key]

    def update_keys(self, keys):
        """ Add new IDs to cache.

            :param list keys: list of new IDs to be added to cache
            :return: self
            :rtype: ObjectCache
        """
        if not self:
            # for large amounts of data, this may be faster (no need for set
            # and difference calls)
            self.update({cid: {'id': cid} for cid in keys})
        else:
            self.update({cid: {'id': cid}
                         for cid in set(keys).difference(six.viewkeys(self))})
        return self

    def update_context(self, new_context):
        """ Updates or sets new context for thes ObjectCache instance

            :param dict new_context: context dictionary to update cached
                                     context with
            :return: updated context
        """
        if new_context is not None:
            if self.context is None:
                self._context = dict(new_context)
            else:
                self._context.update(new_context)
        return self.context

    def get_ids_to_read(self, *fields):
        """ Return list of ids, that have no at least one of specified
            fields in cache

            For example::

                cache.get_ids_to_read('name', 'country_id', 'parent_id')

            This code will traverse all record ids managed by this cache,
            and find those that have no at least one field in cache.
            This is highly useful in prefetching
        """
        return [key for key, val in six.viewitems(self)
                if any(((field not in val) for field in fields))]

    def cache_field(self, rid, ftype, field_name, value):
        """ This method impelment additional caching functionality,
            like caching related fields, and so...

            :param int rid: Record ID
            :param str ftype: field type
            :param str field_name: name of field
            :param value: value to cache for field
        """
        self[rid][field_name] = value
        if value and ftype == 'many2one':
            rcache = self._root_cache[self._object.
                                      columns_info[field_name]['relation']]

            if isinstance(value, six.integer_types):  # pragma: no cover
                # internal dict {'id': key} will be created by default
                # (see ObjectCache.__missing__)
                rcache[value]
            elif isinstance(value, (list, tuple)):
                # usualy for many2one fields odoo returns tuples like
                # (id, name), where id is ID of remote record, and name
                # is human readable name of record (result of name_get method)
                # so we cache this name for futher usage too
                rcache[value[0]]['__name_get_result'] = value[1]
        elif value and ftype in ('many2many', 'one2many'):
            rcache = self._root_cache[self._object.
                                      columns_info[field_name]['relation']]
            rcache.update_keys(value)

    def parse_prefetch_fields(self, fields):
        """ Parse fields to be prefetched, sparating, cache's object fields
            and related fields.

            Used internaly

            :param list fields: list of fields to prefetch
            :return: returns ``tuple(prefetch_fields, related_fields)``,
                where ``prefetch_fields`` is list of fields,
                to be read for current object, and ``related_fields`` is
                dictionary of form:
                ``{'related.object': ['relatedfield1', 'relatedfield2.relatedfield']}``
            :rtype: tuple
        """  # noqa
        rel_fields = collections.defaultdict(list)
        prefetch_fields = set()
        for field in fields:
            field_path = field.split('.', 1)
            xfield = field_path.pop(0)
            xfield_info = self._object.columns_info.get(xfield, None)
            if xfield_info is not None:
                prefetch_fields.add(xfield)
                relation = xfield_info.get('relation', False)
                if field_path and relation:
                    # only one item left
                    rel_fields[relation].append(field_path[0])

        return list(prefetch_fields), rel_fields

    def prefetch_fields(self, fields):
        """ Prefetch specified fields for this cache.
            Also, dot (".") may be used in field name
            to prefetch related fields::

                cache.prefetch_fields(
                    ['myfield1', 'myfields2_ids.relatedfield'])

            :param list fields: list of fields to prefetch

        """
        to_prefetch, related = self.parse_prefetch_fields(fields)

        col_info = self._object.columns_info
        for data in self._object.read(self.get_ids_to_read(*to_prefetch),
                                      to_prefetch):
            for field, value in data.items():

                # Fill related cache
                ftype = col_info.get(field, {}).get('type', None)
                self.cache_field(data['id'], ftype, field, value)

        if related:
            # TODO: think how to avoid infinite recursion and double reads
            for obj_name, rfields in related.items():
                self._root_cache[obj_name].prefetch_fields(rfields)


class Cache(dict):
    """ Cache to be used for Record's data.

        This is root cache, which manages model local cache

        cache['res.partner'] -> ObjectCache('res.partner')
    """
    __slots__ = ('_client',)

    def __init__(self, client, *args, **kwargs):
        self._client = client
        super(Cache, self).__init__(*args, **kwargs)

    @property
    def client(self):
        """ Access to Client instance this cache belongs to
        """
        return self._client

    def __missing__(self, key):
        try:
            obj = self._client.get_obj(key)
        except ValueError:
            raise KeyError("There is no object with such name: %s" % key)

        # TODO: FIX: Object caches generated without context
        self[key] = ObjectCache(self, obj)
        return self[key]


def empty_cache(client):
    """ Create instance of empty cache for Record

        :param Client client: instance of Client to create cache for
        :return: instance of Cache class
        :rtype: Cache

        Cache is dictionary-like object with structure like::

            cache = {
                'product.product': {
                    1: {
                        'id': 1,
                        'name': 'product1',
                        'default_code': 'product1',
                    },
                },
            }

    """
    return Cache(client)



from . import BaseTestCase
from ..client import Client
from ..service.db import to_dbname


import os
import os.path
import unittest
import time
import six


@unittest.skipUnless(os.environ.get('TEST_DB_SERVICE', False),
                     'requires DB service test enabled')
class Test_999_DB(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             protocol=self.env.protocol,
                             port=self.env.port)

    def test_00_create_db(self):
        # create database if it was not created yet
        if not self.client.services.db.db_exist(self.env.dbname):
            self.client.services.db.create_db(self.env.super_password,
                                              self.env.dbname, demo=True,
                                              admin_password=self.env.password)
        else:
            raise unittest.SkipTest("database already creaated. "
                                    "no need to create it")

    def test_10_to_dbname(self):
        cl_db = self.client.login(self.env.dbname, 'admin', self.env.password)

        res = to_dbname(self.env.dbname)
        self.assertEqual(res, self.env.dbname)
        self.assertIsInstance(res, six.string_types)

        res = to_dbname(cl_db)
        self.assertEqual(res, self.env.dbname)
        self.assertIsInstance(res, six.string_types)

        # Test that value error is raise when unexpected value passed to
        # function
        with self.assertRaises(ValueError):
            to_dbname(25)

    def test_20_dump_drop_restore(self):
        # dump db
        dump_data = self.client.services.db.dump_db(self.env.super_password,
                                                    self.env.dbname)
        self.assertIsInstance(dump_data, six.binary_type)

        # drop it
        self.client.services.db.drop_db(self.env.super_password,
                                        self.env.dbname)
        self.assertNotIn(self.env.dbname, self.client.services.db)

        # and try to restore it
        self.client.services.db.restore_db(self.env.super_password,
                                           self.env.dbname,
                                           dump_data)
        self.assertIn(self.env.dbname, self.client.services.db)

        # check if objects were restored
        time.sleep(2)
        cl = self.client.login(self.env.dbname,
                               self.env.user,
                               self.env.password)
        self.assertIsNotNone(cl.uid)
        self.assertIsNotNone(cl.user)



# python imports
import os
import logging
import atexit
import functools

# project imports
from .connection import ConnectorBase
from ..exceptions import ConnectorError
from ..utils import ustr


logger = logging.getLogger(__name__)


class LocalConnectorError(ConnectorError):
    """ Local connector error wrapper
    """
    def __init__(self, exc):
        self.exc = exc
        super(LocalConnectorError, self).__init__(ustr(exc))


class LocalMethod(object):
    """ Odoo method wrapper
    """
    def __init__(self, service, name):
        self.service = service
        self.name = name
        self.odoo = service.odoo

        # bind odoo's methdo to self._method
        if self.odoo._api_v7:
            self._method = functools.partial(
                self.odoo.netsvc.ExportService.getService(
                    self.service.service_name).dispatch,
                self.name)
        else:
            self._method = functools.partial(
                self.odoo.http.dispatch_rpc,
                self.service.service_name,
                self.name)

    def __call__(self, *args):
        try:
            res = self._method(args)
        except Exception as exc:
            raise LocalConnectorError(exc)
        return res


class LocalService(object):
    """ Local Odoo service
    """
    def __init__(self, connection, service_name):
        self.connection = connection
        self.service_name = service_name
        self.odoo = connection.odoo

        # variable to cache methods
        self._methods = {}

    def __getattr__(self, name):
        meth = self._methods.get(name, None)
        if meth is None:
            try:
                meth = LocalMethod(self, name)
            except KeyError as exc:
                raise LocalConnectorError(exc)
            except AttributeError as exc:
                raise LocalConnectorError(exc)
            except Exception as exc:
                raise LocalConnectorError(exc)
            else:
                self._methods[name] = meth

        return meth


class ConnectorLocal(ConnectorBase):
    """ Connect to local odoo instal.

        NOTE: To use this connector, odoo must be importable as 'odoo' or
              'openerp'. This connector will automaticaly determine
              Odoo version, and organize correct bechavior

        NOTE2: This connector tested only on **python2.7** because
               Odoo uses this version of python

        NOTE3: Because, standard params have no sense for this connector,
               it ignores them, but instead, it looks in extra_args
               for argument 'local_args', which must be a list of command_line
               args to run odoo with
        NOTE4: This connector still in alpha stage, so it may introduce some
               bugs
    """
    class Meta:
        name = 'local'

    # Need for backward compatability, because there 'verbose' keyword argument
    # may be present in extra_args due to old sessions saved with this arg
    def __init__(self, *args, **kwargs):
        super(ConnectorLocal, self).__init__(*args, **kwargs)
        self.extra_args.pop('verbose', None)

        self.odoo_args = self.extra_args.get('local_args', [])

        self.odoo = self._start_odoo_services()

    # This code is based on erppeek's code:
    #     https://github.com/tinyerp/erppeek/blob/master/erppeek.py#L251
    def _start_odoo_services(self):
        """Initialize the Odoo services.
           return `odoo` package
        """
        try:
            # Odoo 10.0+
            import odoo
            import odoo.release  # to avoid 9.0 with odoo.py on path
        except ImportError:
            try:
                # Odoo 9.0 and less versions
                import openerp as odoo
            except ImportError:
                raise

        if getattr(odoo, '_odoo_services_started', False):
            # If odoo services already started, just return odoo instance
            return odoo

        if odoo.release.version_info < (7,):
            raise ConnectorError(
                "Unsupported Odoo version: %s" % odoo.release.version_info)

        odoo._api_v7 = odoo.release.version_info < (8,)
        os.putenv('TZ', 'UTC')
        odoo.tools.config.parse_config(self.odoo_args)

        if odoo._api_v7:
            odoo.service.start_internal()
        else:   # Odoo v8
            try:
                odoo.api.Environment._local.environments = \
                    odoo.api.Environments()
            except AttributeError:
                pass

        @atexit.register
        def close_all():
            if odoo.release.version_info < (10,):
                # Odoo 8-9
                RegistryManager = odoo.modules.registry.RegistryManager
            else:
                # Odoo 10+
                RegistryManager = odoo.modules.registry.Registry

            for db in RegistryManager.registries.keys():
                odoo.sql_db.close_db(db)

        # Mark odoo, that it have services started
        odoo._odoo_services_started = True

        return odoo

    def _get_service(self, name):
        return LocalService(self, name)



# python imports
import simplejson  # Standard json cannot dump bytes on py3
import random
import requests
import logging

# project imports
from .connection import ConnectorBase
from .. import exceptions as exceptions
from ..utils import ustr


logger = logging.getLogger(__name__)


class JSONRPCError(exceptions.ConnectorError):
    """ JSON-RPC error wrapper
    """
    def __init__(self, message, code=None, data=None):
        self.message = message
        self.code = code
        self.data = data

        if self.data_message and self.data_debug:
            msg = u"""%(message)s\n%(debug)s\n""" % self.data
        elif self.data:
            msg = ustr(self.data)
        else:
            msg = self.message

        super(JSONRPCError, self).__init__(msg)

    @property
    def data_message(self):
        """ Error message got from Odoo server
        """
        if self.data:
            return self.data.get('message', None)

    @property
    def data_debug(self):
        """ Debug information got from Odoo server

            Usualy traceback
        """
        if self.data:
            return self.data.get('debug', None)


# TODO: think, may be it is a good idea to reimplement this via functions
class JSONRPCMethod(object):
    """ Class that implements RPC call via json-rpc protocol
    """
    __slots__ = ('__method', '__url', '__service', '__rpc_proxy')

    def __init__(self, rpc_proxy, url, service, method):
        self.__method = method
        self.__url = url
        self.__service = service
        self.__rpc_proxy = rpc_proxy

    def prepare_method_data(self, *args):
        """ Prepare data for JSON request
        """
        return {
            "jsonrpc": "2.0",
            "method": 'call',
            "params": {
                "service": self.__service,
                "method": self.__method,
                "args": args,
            },
            "id": random.randint(0, 1000000000),
        }

    def __call__(self, *args):
        method_data = self.prepare_method_data(*args)
        data = simplejson.dumps(method_data)

        # Call rpc
        try:
            res = requests.post(self.__url, data=data, headers={
                "Content-Type": "application/json",
            }, verify=self.__rpc_proxy.ssl_verify)
        except requests.exceptions.RequestException as exc:
            msg = ("Cannot connect to url %s\n"
                   "Exception %s raised!" % (self.__url, exc))
            logger.error(msg)
            raise JSONRPCError(msg)

        # Process results
        try:
            result = simplejson.loads(res.text)
        except ValueError:
            info = {
                "original_url": self.__url,
                "url": res.url,
                "code": res.status_code,
                "content": res.text[:2000],
                "method_data": method_data,
            }
            logger.error("Cannot decode JSON")
            raise JSONRPCError("Cannot decode JSON: %s" % info)

        if result.get("error", None):
            error = result['error']
            raise JSONRPCError(error['message'],
                               code=error.get('code', None),
                               data=error.get('data', None))
        # if 'result' is not present in response object, then it seems, that
        # result is None
        return result.get("result", None)


class JSONRPCProxy(object):
    """ Simple Odoo service proxy wrapper
    """
    def __init__(self, host, port, service, ssl=False, ssl_verify=True):
        self.host = host
        self.port = port
        self.service = service

        addr = host
        if port not in (None, 80):
            addr += ':%s' % self.port
        self.url = '%s://%s/jsonrpc' % (ssl and 'https' or 'http', addr)

        # request parametrs
        self.ssl_verify = ssl_verify

        # variable to cach methods
        self._methods = {}

    def __getattr__(self, name):
        meth = self._methods.get(name, None)
        if meth is None:
            self._methods[name] = meth = JSONRPCMethod(self,
                                                       self.url,
                                                       self.service,
                                                       name)
        return meth


class ConnectorJSONRPC(ConnectorBase):
    """ JSON-RPC connector

        available extra arguments:
            - ssl_verify: (optional) if True, the SSL cert will be verified.
    """
    class Meta:
        name = 'json-rpc'
        use_ssl = False

    # Need for backward compatability, because there 'verbose' keyword argument
    # may be present in extra_args due to old sessions saved with this arg
    def __init__(self, *args, **kwargs):
        super(ConnectorJSONRPC, self).__init__(*args, **kwargs)
        self.extra_args.pop('verbose', None)

    def _get_service(self, name):
        return JSONRPCProxy(self.host,
                            self.port,
                            name,
                            ssl=self.Meta.use_ssl,
                            **self.extra_args)


class ConnectorJSONRPCS(ConnectorJSONRPC):
    """ JSON-RPCS Connector
    """
    class Meta:
        name = 'json-rpcs'
        use_ssl = True



from .cache import empty_cache          # noqa
from .object import (get_object,        # noqa
                     Object)            # noqa
from .record import (get_record,        # noqa
                     get_record_list,   # noqa
                     Record,            # noqa
                     RecordList)        # noqa
from .service import Service            # noqa



""" This module contains classes and logic to handle operations on records
"""

from ..utils import (wpartial,
                     normalizeSField,
                     preprocess_args,
                     ustr,
                     DirMixIn)
from .object import Object
from .cache import (empty_cache,
                    Cache)


import six
import abc
import numbers
import functools
import collections
from extend_me import (ExtensibleType,
                       ExtensibleByHashType)


__all__ = (
    'Record',
    'RecordRelations',
    'ObjectRecords',
    'RecordList',
    'get_record',
    'get_record_list',
)


RecordMeta = ExtensibleByHashType._('Record', hashattr='object_name')


def get_record(obj, rid, cache=None, context=None):
    """ Creates new Record instance

        Use this method to create new records, because of standard
        object creation bypasses extension's magic.

            :param Object obj: instance of Object this record is related to
            :param int rid: ID of database record to fetch data from
            :param cache: Cache instance. (usualy generated by
                          function empty_cache()
            :type cache: Cache
            :param dict context: if specified, then cache's context
                                 will be updated
            :return: created Record instance
            :rtype: Record
    """
    cls = RecordMeta.get_class(obj.name, default=True)
    return cls(obj, rid, cache=cache, context=context)


@six.python_2_unicode_compatible
class Record(six.with_metaclass(RecordMeta, DirMixIn)):
    """ Base class for all Records

        Do not use it to create record instances manualy.
        Use ``get_record`` function instead.
        It implements all extensions mangic

        But class should be used for ``isinstance`` checks.

        It is posible to create extensions of this class that will be binded
        only to specific Odoo objects

        For example, if You need to extend all recrods of products,
        do something like this::

            class MyProductRecord(Record):
                class Meta:
                    object_name = 'product.product'

                def __init__(self, *args, **kwargs):
                    super(MyProductRecord, self).__init__(*args, **kwargs)

                    # to avoid double read, save once read value to record
                    # instance
                    self._sale_orders = None

                @property
                def sale_orders(self):
                    ''' Sale orders related to curent product
                    '''
                    if self._sale_orders is None:
                        so = self._client['sale.order']
                        domain = [('order_line.product_id', '=', self.id)]
                        self._sale_orders = so.search_records(
                                                domain, cache=self._cache)
                    return self._sale_orders

        And atfter this, next code is valid::

            products = client['product.product'].search_records([])
            products_so = products.filter(lambda p: bool(p.sale_orders))
            products_so_gt_10 = products.filter(
                lambda p: len(p.sale_orders) > 10)

            for product in products_so_gt_10:
                print("Product: %s" % product.default_code)
                for pso in product.sale_orders:
                    print("\t%s" % pso.name)


        :param Object obj: instance of object this record is related to
        :param int rid: ID of database record to fetch data from
        :param cache: Cache instance.
                      (usualy generated by function empty_cache())
        :type cache: Cache
        :param dict context: if specified, then cache's context
                             will be updated

        Note, to create instance of cache call *empty_cache*
    """

    __slots__ = ['_object', '_cache', '_lcache', '_id', '_related_objects']

    def __init__(self, obj, rid, cache=None, context=None):
        assert isinstance(obj, Object), "obj should be Object"
        assert isinstance(rid, numbers.Integral), "rid must be int"

        self._id = rid
        self._object = obj
        self._cache = empty_cache(obj.client) if cache is None else cache
        self._lcache = self._cache[obj.name]
        self._related_objects = {}

        self._lcache[self._id]  # ensure that ID of this record is in cache.
        if context is not None:
            self._lcache.update_context(context)

    def __dir__(self):
        res = super(Record, self).__dir__()
        res.extend(self._columns_info.keys())
        res.extend(self._object.stdcall_methods)
        return list(set(res))

    @property
    def id(self):
        """ Record ID

            :rtype: int
        """
        return self._id

    @property
    def _data(self):
        """ Data dictionary for this record.
            (Just a client to cache)

            :rtype: dict
        """
        return self._lcache[self._id]

    @property
    def context(self):
        """ Returns context to be used for thist record
        """
        return self._lcache.context

    @property
    def _service(self):
        """ Returns instance of related Object service instance
        """
        return self._object.service

    @property
    def _client(self):
        """ Returns instance of related Client object

            :rtype: openerp_proxy.core.Client
        """
        return self._object.client

    @property
    def _columns_info(self):
        """ Returns dictionary with information about columns of related object
        """
        return self._object.columns_info

    @property
    def as_dict(self):
        """ Provides dictionary with record's data in raw form

            :rtype: dict
        """
        return self._data.copy()

    @property
    def _name(self):
        """ Returns result of ``name_get`` for this record

            :rtype: str
        """
        if self._data.get('__name_get_result', None) is None:
            lcache = self._lcache
            data = self._object.name_get(list(lcache), context=self.context)
            for _id, name in data:
                lcache[_id]['__name_get_result'] = name
        return self._data.get('__name_get_result', u'ERROR')

    def __str__(self):
        return u"R(%s, %s)[%s]" % (self._object.name,
                                   self.id,
                                   ustr(self._name))

    def __repr__(self):
        return str(self)

    def __int__(self):
        return self._id

    def __hash__(self):
        return hash((self._object.name, self._id))

    def __eq__(self, other):
        if isinstance(other, Record):
            return other.id == self._id

        if isinstance(other, numbers.Integral):
            return self._id == other

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _get_many2one_rel_obj(self, name, rel_data, cached=True):
        """ Method used to fetch related object by name of field
            that points to it
        """
        if name not in self._related_objects or not cached:
            if rel_data:
                # Do not forged about relations in form [id, name]
                rel_id = (rel_data[0]
                          if isinstance(rel_data, collections.Iterable)
                          else rel_data)

                rel_obj = self._service.get_obj(
                    self._columns_info[name]['relation'])
                self._related_objects[name] = get_record(rel_obj,
                                                         rel_id,
                                                         cache=self._cache,
                                                         context=self.context)
            else:
                self._related_objects[name] = False
        return self._related_objects[name]

    def _get_one2many_rel_obj(self, name, rel_ids, cached=True, limit=None):
        """ Method used to fetch related objects by name of field
            that points to them using one2many relation
        """
        if name not in self._related_objects or not cached:
            rel_obj = self._service.get_obj(
                self._columns_info[name]['relation'])
            self._related_objects[name] = get_record_list(rel_obj,
                                                          rel_ids,
                                                          cache=self._cache,
                                                          context=self.context)
        return self._related_objects[name]

    def _get_field(self, ftype, name):
        """ Returns value for field 'name' of type 'type'

            :param str ftype: type of field to det value for
            :param str name: name of field to read

            Should be overridden by extensions to provide better hadling
            for diferent field values
        """
        if name not in self._data:
            # save 'cache_field' function before for loop
            cache_field = self._lcache.cache_field

            # get list of ids in cache, that have not read requested field
            for data in self._object.read(self._lcache.get_ids_to_read(name),
                                          [name],
                                          context=self.context):
                # write each row of data to cache
                cache_field(data['id'], ftype, name, data[name])

        # relational fields
        if ftype == 'many2one':
            return self._get_many2one_rel_obj(name, self._data[name])
        if ftype in ('one2many', 'many2many'):
            return self._get_one2many_rel_obj(name, self._data[name])

        return self._data[name]

    # Allow dictionary access to data fields
    def __getitem__(self, name):
        if name == 'id':
            return self.id

        field = self._columns_info.get(name, None)

        if field is None:
            raise KeyError("No such field %s in object %s, %s"
                           "" % (name, self._object.name, self.id))

        ftype = field and field['type']

        # TODO: refactore to be able to pass field instead of only field type
        return self._get_field(ftype, name)

    # Allow to access data as attributes and call object's methods
    # directly from record object
    def __getattr__(self, name):
        try:
            res = self[name]   # Try to get data field
        except KeyError:
            method = getattr(self._object, name)
            res = wpartial(method, [self.id])
            setattr(self, name, res)
        return res

    def refresh(self):
        """Reread data and clean-up the caches

           :returns: self
           :rtype: Record
        """
        self._data.clear()
        self._data['id'] = self._id

        # Update related objects cache
        rel_objects = self._related_objects
        self._related_objects = {}  # cleanup related_objects cache

        # recursively cleanup related records
        for rel in rel_objects.values():
            if isinstance(rel, (Record, RecordList)):
                # both, Record and RecordList objects have *refresh* method
                rel.refresh()

        return self

    def read(self, fields=None, context=None, multi=False):
        """ Rereads data for this record (or for al records in whole cache)

            :param list fields: list of fields to be read (optional)
            :param dict context: context to be passed to read (optional)
                                 does not midify record's context
            :param bool multi: if set to True, that data will be read for
                               all records of this object in current
                               cache (query).
            :return: dict with data had been read
            :rtype: dict
        """
        ctx = {} if self.context is None else self.context.copy()
        if context is not None:
            ctx.update(context)

        ids = list(self._lcache) if multi else [self.id]
        args, kwargs = preprocess_args(ids, fields, context=ctx or None)

        res = {}
        for rdata in self._object.read(*args, **kwargs):
            self._lcache[rdata['id']].update(rdata)
            if rdata['id'] == self.id:
                res = rdata
        return res

    def copy(self, default=None, context=None):
        """ copy this record.

            :param dict default: dictionary default values for new record
                                 (optional)
            :param dict context: dictionary with context used to copy
                                 this record. (optional)
            :return: Record instance for created record
            :rtype: Record

            Note about context: by default cache's context will be used,
            and if some context will be passed to this method, new dict,
            which is combination of default context and passed context,
            will be passed to server.
        """
        ctx = {} if self.context is None else self.context.copy()

        if context is not None:
            ctx.update(context)

        # None values should not be passed via xml-rpc
        args, kwargs = preprocess_args(self.id,
                                       default=default,
                                       context=ctx or None)

        new_id = self._object.copy(*args, **kwargs)
        return get_record(self._object,
                          new_id,
                          cache=self._cache,
                          context=self.context)

    def get(self, field_name, default=None):
        """ Try to get field *field_name*, if if field name is not available
            return *default* value for it

            if *default* is None and it is not possible to get field value,
            then raises *KeyErro*

            :param str field_name: name of field to get value for
            :param default: default value for case when no such field
            :return: field value
            :raises KeyError: if cannot get field value

            Note: This may be useful for code that expected to be working for
            different Odoo versions which have different database schemes.
        """
        try:
            res = self[field_name]
        except KeyError:
            if default is None:
                raise
            else:
                res = default
        return res


RecordListMeta = ExtensibleType._('RecordList', with_meta=abc.ABCMeta)


def get_record_list(obj, ids=None, fields=None, cache=None, context=None):
    """ Returns new instance of RecordList object.

        :param obj: instance of Object to make this list related to
        :type obj: Object
        :param ids: list of IDs of objects to read data from
        :type ids: list of int
        :param fields: list of field names to read by default  (not used now)
        :type fields: list of strings (not used now)
        :param cache: Cache instance.
                      (usualy generated by function empty_cache()
        :type cache: Cache
        :param context: context to be passed automatically to methods
                        called from this list (not used yet)
        :type context: dict
    """
    return RecordListMeta.get_object(obj,
                                     ids,
                                     fields=fields,
                                     cache=cache,
                                     context=context)


# TODO:  impelment additional operators
#    - operator: +
#    - operator: +=
#
# TODO: implement correct bechavior of cache when adding new records to record
# list with diferent cache
@six.python_2_unicode_compatible
class RecordList(six.with_metaclass(RecordListMeta,
                                    collections.MutableSequence,
                                    DirMixIn)):
    """Class to hold list of records with some extra functionality

        :param obj: instance of Object to make this list related to
        :type obj: Object
        :param ids: list of IDs of objects to read data from
        :type ids: list of int
        :param fields: list of field names to read by default
        :type fields: list of strings
        :param cache: Cache instance.
                      (usualy generated by function empty_cache()
        :type cache: Cache
        :param context: context to be passed automatically to methods
                        called from this list (not used yet)
        :type context: dict

    """
    __slots__ = ('_object', '_cache', '_lcache', '_records')

    def __init__(self, obj, ids=None, fields=None, cache=None, context=None):
        """
        """
        self._object = obj
        self._cache = empty_cache(obj.client) if cache is None else cache
        self._lcache = self._cache[obj.name]

        if context is not None:
            self._lcache.update_context(context)

        ids = [] if ids is None else ids

        # We need to add these ids to cache to make prefetching and data
        # reading work correctly. if some of ids will not be present in cache,
        # then, on access to field of record with such id, data will not be
        # read from database.
        # Look into *Record._get_field* method for more info
        self._lcache.update_keys(ids)

        _cache = self._cache  # before loop, save cache in separate variable
        self._records = [get_record(obj, id_, cache=_cache)
                         for id_ in ids]

        # if there some fields prefetching was requested, do it
        if fields is not None:
            self.prefetch(*fields)

    def __dir__(self):
        res = super(RecordList, self).__dir__()
        res.extend(self._object.stdcall_methods)
        return list(set(res))

    @property
    def object(self):
        """ Object this record is related to
        """
        return self._object

    @property
    def context(self):
        """ Returns context to be used for this list
        """
        return self._lcache.context

    @property
    def ids(self):
        """ IDs of records present in this RecordList
        """
        return [r.id for r in self._records]

    @property
    def records(self):
        """ Returns list (class 'list') of records
        """
        return self._records

    @property
    def length(self):
        """ Returns length of this record list
        """
        return len(self._records)

    def _new_context(self, new_context=None):
        """ Create new context which is combination of *self.context*
            and passed context argument.

            mostly for internal usage

            :param dict new_context: new context. default is None
            :return: new context dict which is combination of
                     *self.context* and *new_context* or *None*
            :rtype: dict|None
        """
        if new_context is None:
            return self.context

        ctx = {} if self.context is None else self.context.copy()
        ctx.update(new_context)
        return ctx

    # Container related methods
    def __getitem__(self, index):
        if isinstance(index, slice):
            # Note no context passed, because it is stored in cache
            return get_record_list(self.object,
                                   ids=[r.id for r in self._records[index]],
                                   cache=self._cache)
        return self._records[index]

    def __setitem__(self, index, value):
        if isinstance(value, Record):
            self._records[index] = value
        else:
            raise ValueError("In 'RecordList[index] = value' operation, "
                             "value must be instance of Record")

    def __delitem__(self, index):
        del self._records[index]

    def __iter__(self):
        return iter(self._records)

    def __len__(self):
        return self.length

    def __contains__(self, item):
        if isinstance(item, numbers.Integral):
            return item in self.ids
        if isinstance(item, Record):
            return item in self._records
        return False

    def insert(self, index, item):
        """ Insert record to list

            :param item: Record instance to be inserted into list.
                         if int passed, it considered to be ID of record
            :type item: Record|int
            :param int index: position where to place new element
            :return: self
            :rtype: RecordList
        """
        assert isinstance(item, (Record, numbers.Integral)), \
            "Only Record or int instances could be added to list"
        if isinstance(item, Record):
            self._records.insert(index, item)
        else:
            self._records.insert(index, self._object.read_records(
                item, cache=self._cache))
        return self

    # Overridden to make ability to call methods of object on list of IDs
    # present in this RecordList
    def __getattr__(self, name):
        method = getattr(self.object, name)
        kwargs = {} if self.context is None else {'context': self.context}
        res = wpartial(method, self.ids, **kwargs)
        return res

    def __str__(self):
        return u"RecordList(%s): length=%s" % (self.object.name, self.length)

    def __repr__(self):
        return str(self)

    def refresh(self):
        """ Cleanup data caches. next try to get data will cause rereading of it

           :returns: self
           :rtype: instance of RecordList
        """
        for record in self.records:
            record.refresh()
        return self

    def sort(self, key=None, reverse=False):
        """ sort(key=None, reverse=False) -- inplace sort

            anyfield.SField instances may be safely passed as 'key' arguments.
            no need to convert them to function explicitly

            :return: self
        """
        if callable(key):
            key = normalizeSField(key)

        self._records.sort(key=key, reverse=reverse)
        return self

    def group_by(self, grouper):
        """ Groups all records in list by specifed grouper.

            :param grouper: field name or callable to group results by.
                            if callable is passed, it should receive only
                            one argument - record instance, and result of
                            calling grouper will be used as key
                            to group records by.
            :type grouper: string|callable(record)|anyfield.SField
            :return: dictionary

            for example we have list of sale orders and
            want to group it by state

            .. code-block:: python

              # so_list - variable that contains list of sale orders selected
              # by some criterias. so to group it by state we will do:
              group = so_list.group_by('state')

              # Iterate over resulting dictionary
              for state, rlist in group.iteritems():
                  # Print state and amount of items with such state
                  print state, rlist.length

            or imagine that we would like to group records
            by last letter of sale order number

            .. code-block:: python

              # so_list - variable that contains list of sale orders selected
              # by some criterias. so to group it by last letter of sale
              # order name  we will do:
              group = so_list.group_by(lambda so: so.name[-1])

              # Iterate over resulting dictionary
              for letter, rlist in group.iteritems():
                  # Print state and amount of items with such state
                  print letter, rlist.length

        """
        if callable(grouper):
            grouper = normalizeSField(grouper)

        cls_init = functools.partial(get_record_list,
                                     self.object,
                                     ids=[],
                                     cache=self._cache)
        res = collections.defaultdict(cls_init)
        for record in self.records:
            if isinstance(grouper, six.string_types):
                key = record[grouper]
            elif callable(grouper):
                key = grouper(record)

            res[key].append(record)
        return res

    def filter(self, func):
        """ Filters items using *func*.

            :param func: callable to check if record should be included
                         in result.
            :type func: callable(record)->bool|anyfield.SField
            :return: RecordList which contains records that matches results
            :rtype: RecordList
        """
        func = normalizeSField(func)
        return get_record_list(self.object,
                               ids=[r.id for r in self.records if func(r)],
                               cache=self._cache)

    def mapped(self, field):
        """ **Experimental**, Provides similar functionality
            to Odoo's mapped() method, but supports only dot-separated
            field name as argument, no callables yet.

            Returns list of values of field of each record in this recordlist.
            If value of field is RecordList or Record instance,
            than RecordList instance will be returned

            Thus folowing code will work

            .. code:: python

                # returns a list of names
                records.mapped('name')

                # returns a recordset of partners
                record.mapped('partner_id')

                # returns the union of all partner banks,
                # with duplicates removed
                record.mapped('partner_id.bank_ids')

            :param str field: returns list of values of 'field'
                              for each record in this RecordList
            :rtype: list or RecordList
        """
        def get_field(rec):
            fields = field.split('.')
            val = rec
            while fields and val:
                f = fields.pop(0)
                val = val[f]

            return val

        # Choose type of result
        (res_model,
         res_field,
         res_rel_model) = self._object.resolve_field_path(field)[-1]
        if res_rel_model:
            res_obj = self._object.client[res_rel_model]
            res = get_record_list(res_obj,
                                  [],
                                  cache=self._cache,
                                  context=self.context)
        else:
            res = []

        for record in self.records:
            val = get_field(record)
            if not val:
                continue

            if isinstance(val, RecordList):
                res.extend(val)
            elif val not in res:
                res.append(val)

        return res

    def copy(self, context=None, new_cache=False):
        """ Returns copy of this list, possibly with modified context
            and new empty cache.

            :param dict context: new context values to be used on new list
            :param bool new_cache: if set to True, then new cache instance
                                   will be created for resulting recordlist
                                   if set to Cache instance, than it will be
                                   used for resulting recordlist
            :return: copy of this record list.
            :rtype: RecordList
            :raises ValueError: when incorrect value passed to new_cache
        """
        if isinstance(new_cache, Cache):
            cache = new_cache
        elif not new_cache:
            cache = self._cache
        elif new_cache is True:
            cache = empty_cache(self.object.client)
        else:
            raise ValueError("Wrong value for parametr 'new_cache': %r"
                             "" % (new_cache,))

        return get_record_list(self.object,
                               ids=self.ids,
                               cache=cache,
                               context=context)

    def existing(self, uniqify=True):
        """ Filters this list with only existing items

            :parm bool uniqify: if set to True, then all dublicates
                                will be removed. Default: True
            :return: new RecordList instance
            :rtype: RecordList
        """
        existing_ids = self.exists()
        new_ids = []
        for id_ in self.ids:
            if id_ not in existing_ids:
                continue
            if uniqify and id_ in new_ids:
                continue
            new_ids.append(id_)
        return get_record_list(self.object,
                               ids=new_ids,
                               cache=self._cache)

    def prefetch(self, *fields):
        """ Prefetches specified fields into cache
            if no fields passed, then all 'simple_fields' will be prefetched

            By default field read performed only when that field is requested,
            thus when You need to read more then one field, few rpc requests
            will be performed. to avoid multiple unneccessary rpc calls this
            method is implemented.

            :return: self, which allows chaining of operations
            :rtype: RecordList
        """
        fields = fields if fields else self.object.simple_fields

        self._lcache.prefetch_fields(fields)

        return self

    # remote method overrides
    def search(self, domain, *args, **kwargs):
        """ Performs normal search, but adds ``('id', 'in', self.ids)``
            to search domain

            :returns: list of IDs found
            :rtype: list of integers
        """
        ctx = self._new_context(kwargs.get('context', None))

        if ctx is not None:
            kwargs['context'] = ctx

        return self.object.search([('id', 'in', self.ids)] + domain,
                                  *args,
                                  **kwargs)

    def search_records(self, domain, *args, **kwargs):
        """ Performs normal search_records, but adds
            ``('id', 'in', self.ids)`` to domain

            :returns: RecordList of records found
            :rtype: RecordList instance
        """
        ctx = self._new_context(kwargs.get('context', None))

        if ctx is not None:
            kwargs['context'] = ctx

        return self.object.search_records([('id', 'in', self.ids)] + domain,
                                          *args,
                                          **kwargs)

    def read(self, fields=None, context=None):
        """ Read wrapper. Takes care about adding RecordList's context to
            object's read method.

            **Warning**: does not update cache by data been read
        """
        ctx = self._new_context(context)
        args, kwargs = preprocess_args(fields, context=ctx)
        return self.object.read(self.ids, *args, **kwargs)


# For backward compatability
RecordRelations = Record


class ObjectRecords(Object):
    """ Adds support to use records from Object classes
    """

    def __init__(self, *args, **kwargs):
        super(ObjectRecords, self).__init__(*args, **kwargs)
        self._model = None

    @property
    def model(self):
        """ Returns Record instance of model related to this object.
            Useful to get additional info on object.
        """
        if self._model is None:
            model_obj = self.client.get_obj('ir.model')
            res = model_obj.search_records([('model', '=', self.name)],
                                           limit=2)
            assert res.length == 1, \
                "There must be only one model for this name"
            self._model = res[0]

        return self._model

    @property
    def model_name(self):
        """ Result of name_get called on object's model
        """
        return self.model._name

    @property
    def simple_fields(self):
        """ List of simple fields which could be fetched fast enough

            This list contains all fields that are not function nor binary

            :type: list of strings
        """
        return [f for f, d in six.iteritems(self.columns_info)
                if d['type'] != 'binary' and not d.get('function', False)]

    def search_records(self, *args, **kwargs):
        """ Return instance or list of instances of Record class,
            making available to work with data simpler

            :param domain: list of tuples, specifying search domain
            :param int offset: (optional) number of results to skip
                               in the returned values
                               (default:0)
            :param limit: optional max number of records  in result
                          (default: False)
            :type limit: int|False
            :param order: optional columns to sort
            :type order: str
            :param dict context: optional context to pass to *search* method
            :param count: if set to True, then only amount of recrods found
                          will be returned.
                          (default: False)
            :param read_fields: optional. specifies list of fields to read.
            :type read_fields: list of strings
            :param Cache cache: cache to be used for records and recordlists
            :return: RecordList contains records found, or integer
                     that represents amount of records found (if count=True)
            :rtype: RecordList|int

            For example:

            .. code:: python

                >>> so_obj = db['sale.order']
                >>> data = so_obj.search_records([('date','>=','2013-01-01')])
                >>> for order in data:
                ...     order.write({'note': 'order date is %s'%order.date})
        """

        # TODO: use search_read for odoo versions >= 8.0
        read_fields = kwargs.pop('read_fields', None)
        cache = kwargs.pop('cache', None)
        context = kwargs.get('context', None)

        if kwargs.get('count', False):
            return self.search(*args, **kwargs)

        res = self.search(*args, **kwargs)
        if not res:
            return get_record_list(self,
                                   ids=[],
                                   fields=read_fields,
                                   context=context,
                                   cache=cache)

        if read_fields:
            return self.read_records(res,
                                     read_fields,
                                     context=context,
                                     cache=cache)
        return self.read_records(res, context=context, cache=cache)

    def read_records(self, ids, fields=None, context=None, cache=None):
        """ Return instance or RecordList class,
            making available to work with data simpler

            :param ids: ID or list of IDS to read data for
            :type ids: int|list of int
            :param list fields: list of fields to read (*optional*)
            :param dict context: context to be passed to read. default=None
            :param Cache cache: cache to use for records and record lists.
                          Pass None to create new cache. default=None.
            :return: Record instance if *ids* is int or RecordList instance
                     if *ids* is list of ints
            :rtype: Record|RecordList

            For example:

            .. code:: python

                >>> so_obj = db['sale.order']
                >>> data = so_obj.read_records([1,2,3,4,5])
                >>> for order in data:
                        order.write({'note': 'order data is %s'%order.data})
        """
        if isinstance(ids, numbers.Integral):
            record = get_record(self, ids, context=context)
            if fields is not None:
                record.read(fields)  # read specified fields
            return record
        if isinstance(ids, collections.Iterable):
            return get_record_list(self, ids, fields=fields, context=context)

        raise ValueError("Wrong type for ids argument: %s" % type(ids))

    def browse(self, *args, **kwargs):
        """ Aliase to *read_records* method.
            In most cases same as serverside *browse*
            (i mean server version 7.0)
        """
        return self.read_records(*args, **kwargs)



import six
import os.path
from pkg_resources import parse_version as V

from . import BaseTestCase
from ..client import Client
from ..service.report import (Report,
                              ReportResult)


class Test_Service_Report(BaseTestCase):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.client = Client(self.env.host,
                             dbname=self.env.dbname,
                             user=self.env.user,
                             pwd=self.env.password,
                             protocol=self.env.protocol,
                             port=self.env.port)
        if self.client.server_version == V('7.0'):
            self.report_name = 'sale.order'
        else:  # >7.0
            self.report_name = 'sale.report_saleorder'

    def test_available_reports(self):
        all_reports = self.client['ir.actions.report.xml'].search_records([])
        expected_names = all_reports.mapped('report_name')

        self.assertItemsEqual(
            list(self.client.services.report.available_reports.keys()),
            expected_names)

        for report in self.client.services.report.available_reports.values():
            self.assertIsInstance(report, Report)

    def test_contains(self):
        self.assertIn(self.report_name, self.client.services.report)

    def test_getitem(self):
        self.assertIs(
            self.client.services.report.available_reports[self.report_name],
            self.client.services.report[self.report_name])

        with self.assertRaises(KeyError):
            self.client.services.report['some_unexistent_report']

    def test_getattr(self):
        self.assertIs(
            self.client.services.report.available_reports[self.report_name],
            getattr(self.client.services.report, self.report_name))

        with self.assertRaises(AttributeError):
            getattr(self.client.services.report, 'some_unexistent_report')

    def test_report_report_name(self):
        self.assertEqual(
            self.client.services.report[self.report_name].name,
            self.report_name)

    def test_report_generate(self):
        so = self.client['sale.order'].search_records([], limit=1)[0]

        result = self.client.services.report[self.report_name].generate(so)

        self.assertIsInstance(result, ReportResult)
        self.assertTrue(result.state)
        self.assertIsInstance(result.result, six.binary_type)
        self.assertEqual(result.format, 'pdf')
        self.assertIsInstance(result.content, six.binary_type)

        # save to default path
        self.assertFalse(os.path.exists(result.path))
        result.save()
        self.assertTrue(os.path.exists(result.path))

        # save to specified path
        my_path = './my-test-report.pdf'
        self.assertFalse(os.path.exists(my_path))
        result.save(my_path)
        self.assertTrue(os.path.exists(my_path))



# -*- encoding: utf-8 -*-
import six
import functools

__all__ = ('ustr',
           'AttrDict',
           'DirMixIn',
           'UConverter',
           'wpartial',
           )

# Check if anyfield is installed
# and import function which converts SField instances to functions
try:
    from anyfield import toFn as normalizeSField
except ImportError:
    def normalizeSField(fn):
        return fn


def wpartial(func, *args, **kwargs):
    """Wrapped partial, same as functools.partial decorator,
       but also calls functools.wrap on its result thus shwing correct
       function name and representation.
    """
    partial = functools.partial(func, *args, **kwargs)

    return functools.wraps(func)(partial)


def preprocess_args(*args, **kwargs):
    """ Skip all args, and kwargs that set to None

        Mostly for internal usage.

        Used to workaround xmlrpc None restrictions
    """
    kwargs = {key: val for key, val in kwargs.items() if val is not None}

    # TODO: review this! It may bring errors
    xargs = list(args[:])
    while xargs and xargs[-1] is None:
        xargs.pop()
    return xargs, kwargs


def stdcall(fn):
    """ Simple decorator for server methods, that supports standard call

        If method supports call like
        ``method(ids, <args>, context=context, <kwargs>)``,
        then it may be decrated by this decorator to appear in
        dir(record) and dir(recordlist) calls, thus making it available
        for autocompletition in ipython or other python shells
    """
    fn.__x_stdcall__ = True
    return fn


class UConverter(object):
    """ Simple converter to unicode

        Create instance with specified list of encodings to be used to
        try to convert value to unicode

        Example::

            ustr = UConverter(['utf-8', 'cp-1251'])
            my_unicode_str = ustr(b'hello - привет')
    """
    default_encodings = ['utf-8', 'ascii']

    def __init__(self, hint_encodings=None):
        if hint_encodings:
            self.encodings = hint_encodings
        else:
            self.encodings = self.default_encodings[:]

    def __call__(self, value):
        """ Convert value to unicode

        :param value: the value to convert
        :raise: UnicodeError if value cannot be coerced to unicode
        :return: unicode string representing the given value
        """
        # it is unicode
        if isinstance(value, six.text_type):
            return value

        # it is not binary type (str for python2 and bytes for python3)
        if not isinstance(value, six.binary_type):
            try:
                value = six.text_type(value)
            except Exception:
                # Cannot directly convert to unicode. So let's try to convert
                # to binary, and that try diferent encoding to it
                try:
                    value = six.binary_type(value)
                except:
                    raise UnicodeError('unable to convert to unicode %r'
                                       '' % (value,))
            else:
                return value

        # value is binary type (str for python2 and bytes for python3)
        for ln in self.encodings:
            try:
                res = six.text_type(value, ln)
            except Exception:
                pass
            else:
                return res

        raise UnicodeError('unable to convert to unicode %r' % (value,))


# default converter instance
ustr = UConverter()


# DirMixIn class implementation. To be able to use super calls to __dir__ in
# subclasses (Py 2/3 support)
# code is based on
# http://www.quora.com/How-dir-is-implemented-Is-there-any-PEP-related-to-that
try:
    object.__dir__
except AttributeError:
    class DirMixIn(object):
        """ Mix-in to make implementing __dir__ method in subclasses simpler
        """
        def __dir__(self):
            def get_attrs(obj):
                import types
                if not hasattr(obj, '__dict__'):
                    return []  # slots only
                if not isinstance(obj.__dict__, (dict, types.DictProxyType)):
                    raise TypeError("%s.__dict__ is not a dictionary"
                                    "" % obj.__name__)
                return obj.__dict__.keys()

            def dir2(obj):
                attrs = set()
                if not hasattr(obj, '__bases__'):
                    # obj is an instance
                    if not hasattr(obj, '__class__'):
                        # slots
                        return sorted(get_attrs(obj))
                    klass = obj.__class__
                    attrs.update(get_attrs(klass))
                else:
                    # obj is a class
                    klass = obj

                for cls in klass.__bases__:
                    attrs.update(get_attrs(cls))
                    attrs.update(dir2(cls))
                attrs.update(get_attrs(obj))
                return list(attrs)

            return dir2(self)
else:
    # There are no need to implement any aditional logic for Python 3.3+,
    # because there base class 'object' already have implemented
    # '__dir__' method, which could be accessed via super() by subclasses
    class DirMixIn:
        pass


class AttrDict(dict, DirMixIn):
    """ Simple class to make dictionary able to use attribute get operation
        to get elements it contains using syntax like:

        >>> d = AttrDict(arg1=1, arg2='hello')
        >>> print(d.arg1)
            1
        >>> print(d.arg2)
            hello
        >>> print(d['arg2'])
            hello
        >>> print(d['arg1'])
            1
    """
    def __getattr__(self, name):
        res = None
        try:
            res = super(AttrDict, self).__getitem__(name)
        except KeyError as e:
            raise AttributeError(str(e))
        return res

    def __dir__(self):
        res = super(AttrDict, self).__dir__() + list(self.keys())
        return list(set(res))



""" Report printing logic

Best way to generate report is::

    data_records = client['res.partner'].search_records([], limit=10)
    report = client.services.report['res.partner'].generate(data_records)
    report.content

Or if it is desired to save it on disk::

    data_records = client['res.partner'].search_records([], limit=10)
    report = client.services.report['res.partner'].generate(data_records)
    report.save('filename to save report with')


where *report* is instance of *ReportResult* and *report.content*
returns already *base64* decoded content of report,
which could be directly written to file (or
just use *report.save(path)* method)
"""

import numbers
from extend_me import Extensible

from .service import ServiceBase
from ..orm import (Record,
                   RecordList)

from ..exceptions import ReportError


class ReportResult(Extensible):
    """ Just a simple and extensible wrapper on report result

        As variant of usage - wrap result returned by server methods
        ``report_get`` and ``render_report`` like::

            ReportResult(report_get(report_id))

    """

    def __init__(self, report, result, path=None):
        self._report = report
        self._orig_result = result
        self._state = result.get('state', False)
        self._result = result.get('result', None)
        self._format = result.get('format', None)
        self._content = None
        self._path = path

        # fix result
        if self._result:
            self._result = self._result.encode('utf-8')

    @property
    def state(self):
        """ Result status. only if True, other fields are available
        """
        return self._state

    @property
    def result(self):
        """ Base64-encoded report content.
            To get already decoded report content, use ``.content`` property

            :raises ReportError: When .state property is False.
                                 This may appear in case when report
                                 is not ready yet, when using
                                 *report* and *report_get* methods
        """
        if self.state is False:
            raise ReportError("Report seems to be not ready yet")
        return self._result

    @property
    def format(self):
        """ Report format
        """
        return self._format

    @property
    def content(self):
        """ Report file content. Already base64-decoded
        """
        if self._content is None:
            import base64
            self._content = base64.b64decode(self.result)
        return self._content

    @property
    def path(self):
        """ Path where file is located or will be located on save
        """
        if self._path is None:
            import hashlib
            content_hash = hashlib.sha256(self.content)
            content_hash = content_hash.hexdigest().encode('utf-8')
            report_name_base = self._report.report_action.name.encode('utf-8')
            report_name_base = report_name_base.replace(b'/', b'-')\
                                               .replace(b':', b'-')
            self._path = str(report_name_base +
                             content_hash +
                             b'.' + self.format.encode('utf-8'))
        return self._path

    def save(self, path=None):
        """ Save's file by specified path or if no path specified
            save it in temp dir with automaticly generated name.
        """
        if path is not None:
            self._path = path
        with open(self.path, 'wb') as f:
            f.write(self.content)
        return self


class Report(Extensible):
    """ Class that represents report.

        useful to simplify report generation

        :param ReportService service: instance of report service
                                      to bind report to
        :param Record report: model of report action

    """

    def __init__(self, service, report):
        self._service = service
        self._report = report

    @property
    def service(self):
        """ Service this report is binded to
        """
        return self._service

    @property
    def report_action(self):
        """ Action of this report
        """
        return self._report

    @property
    def name(self):
        """ Name of report
        """
        return self.report_action.report_name

    def generate(self, model_data, report_type='pdf', context=None):
        """ Generate report

            :param model_data: RecordList or Record or list of obj_ids.
                               represent document or documents
                               to generate report for
            :param str report_type: Type of report to generate.
                                    default is 'pdf'.
            :param dict context: Aditional info. Optional.
            :raises: ReportError
            :return: ReportResult instance that contains generated report
            :rtype: ReportResult
        """
        return self.service.generate_report(self.name,
                                            model_data,
                                            report_type=report_type,
                                            context=context)


class ReportService(ServiceBase):
    """ Service class to simplify interaction with 'report' service
    """
    class Meta:
        name = 'report'

    def __init__(self, *args, **kwargs):
        super(ReportService, self).__init__(*args, **kwargs)
        self._reports = None

    def _get_available_reports(self):
        """ Returns list of reports registered in system
        """
        report_obj = self.client.get_obj('ir.actions.report.xml')
        return {r.report_name: Report(self, r)
                for r in report_obj.search_records([])}

    @property
    def available_reports(self):
        """ Returns dictionary with all available reports

            {<report name> : <Report instance>}
        """
        if self._reports is None:
            self._reports = self._get_available_reports()
        return self._reports

    def _prepare_report_data(self, model, ids, report_type):
        """ Performs preparation of data

            :param str model: model name to generate report for
            :param ids: ID or list of IDs to generate report for
            :param str report_type: Type of report.
        """
        ids = [ids] if isinstance(ids, numbers.Integral) else ids
        return {
            'model': model,
            'id': ids[0],
            'ids': ids,
            'report_type': report_type,
        }

    def __getitem__(self, name):
        return self.available_reports[name]

    def __getattr__(self, name):
        try:
            res = self[name]
        except KeyError as exc:
            raise AttributeError(str(exc))
        return res

    def __contains__(self, report):
        return report in self.available_reports

    def report(self, report_name, model, ids, report_type='pdf', context=None):
        """ Proxy to report service *report* method

            :param str report_name: string representing name of report service
            :param str model: name of model to generate report for
            :param ids: list of object ID to get report for (or just single id)
            :type ids: list of int | int
            :param str report_type: Type of report to generate.
                                    default is 'pdf'.
            :param dict context: Aditional info. Optional.
            :return: ID of report to get by method *report_get*
            :rtype: int
        """
        context = {} if context is None else context
        ids = [ids] if isinstance(ids, numbers.Integral) else ids
        data = self._prepare_report_data(model, ids, report_type)
        return self._service.report(self.client.dbname,
                                    self.client.uid,
                                    self.client._pwd,
                                    report_name,
                                    ids,
                                    data,
                                    context)

    def report_get(self, report_id):
        """ Proxy method to report service *report_get* method

            :param int report_id: int that represents ID of report to get
                                  (value returned by report method)
            :return: dictinary with keys:

                     - 'state': boolean, True if report generated correctly
                     - 'result': base64 encoded content of report file
                     - 'format': string representing format,
                       report generated in

            :rtype: dict
        """
        return self._service.report_get(self.client.dbname,
                                        self.client.uid,
                                        self.client._pwd,
                                        report_id)

    def render_report(self, report_name, model, ids, report_type='pdf',
                      context=None):
        """ Proxy to report service *render_report* method

            NOTE: available after version 6.1.

            :param str report_name: string representing name of report service
            :param str model: name of model to generate report for
            :param ids: list of object ID to get report for (or just single id)
            :type ids: list of int | int
            :param str report_type: Type of report to generate.
                                    default is 'pdf'.
            :param dict context: Aditional info. Optional.
            :return: dictinary with keys:
                        - 'state': boolean, True if report generated correctly
                        - 'result': base64 encoded content of report file
                        - 'format': string representing report format

            :rtype: dict
        """
        context = {} if context is None else context
        ids = [ids] if isinstance(ids, numbers.Integral) else ids
        data = self._prepare_report_data(model, ids, report_type)

        return self._service.render_report(self.client.dbname,
                                           self.client.uid,
                                           self.client._pwd,
                                           report_name,
                                           ids,
                                           data,
                                           context)

    def generate_report(self, report_name, report_data, report_type='pdf',
                        context=None):
        """ Generate specified report for specifed report data.
            Report data could be RecordList or Record instance.
            Result is wrapped into ReportResult class


            :param str report_name: string representing name of report service
            :param report_data: RecordList or Record or ('model_name', obj_ids)
                                represent document or documents
                                to generate report for
            :param str report_type: Type of report to generate.
                                    default is 'pdf'.
            :param dict context: Aditional info. Optional.
            :raises: ReportError
            :return: ReportResult instance that contains generated report
            :rtype: ReportResult
        """
        if isinstance(report_data, RecordList):
            obj_ids = report_data.ids
        elif isinstance(report_data, Record):
            obj_ids = [report_data.id]
        else:  # report_data is list of object ids
            obj_ids = report_data

        report_model = self[report_name].report_action.model

        report_result = self.render_report(report_name,
                                           report_model,
                                           obj_ids,
                                           report_type=report_type,
                                           context=context)

        return ReportResult(self.available_reports[report_name],
                            report_result)



class Error(Exception):
    """ Base class for exceptions"""
    pass


class ConnectorError(Error):
    """ Base class for exceptions related to connectors """
    pass


class ClientException(Error):
    """ Base class for client related exceptions
    """
    pass


class ReportError(Error):
    """ Error raise in process of report generation
    """
    pass


class LoginException(ClientException):
    """ This exception should be raised, when operations requires
        login and password. For example interaction with Odoo object service.
    """
    pass


class ObjectException(ClientException):
    """ Base class for exceptions related to Objects """
    pass
